
package Cpp::CountLoop;

use strict;
use warnings;

use Erreurs;
use Cpp::ParserIdents;
use Lib::NodeUtil;
use Cpp::CppNode;
use Cpp::ParseCpp;
use GreenIT::lib::KeepPortalData;

my $rulesDescription = GreenIT::lib::KeepPortalData::keepPortalData();

my $DEBUG = 0;

my $mnemo_OverFlowingLoopCounter = Ident::Alias_OverFlowingLoopCounter();
my $ConcatInLoop__mnemo = Ident::Alias_ConcatInLoop();
my $InstantiationInLoop__mnemo = Ident::Alias_InstantiationInLoop();
my $FunctionCallInLoopTest__mnemo = Ident::Alias_FunctionCallInLoopTest();
my $OpenCloseFileInLoop__mnemo = Ident::Alias_OpenCloseFileInLoop();
my $LoopCondComparedToNonZero__mnemo = Ident::Alias_LoopCondComparedToNonZero();
my $ToManyNestedLoop__mnemo = Ident::Alias_ToManyNestedLoop();

my $nb_OverFlowingLoopCounter = 0;
my $nb_ConcatInLoop = 0;
my $nb_InstantiationInLoop = 0;
my $nb_FunctionCallInLoopTest = 0;
my $nb_OpenCloseFileInLoop = 0;
my $nb_LoopCondComparedToNonZero = 0;
my $nb_ToManyNestedLoop = 0;

sub checkLoopCondition($) {
	my $condNode = shift;
	
  	if ( GetKind ($condNode) eq ConditionKind ) {            
        my $cond = Lib::NodeUtil::GetKindData($condNode)->{'cond'};
        
        # detect a function call
        while ( $cond =~ /(\w+)\s*\(/g ) {
            $nb_FunctionCallInLoopTest++;
			my $numline = $condNode->[5];
			CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_FunctionCallInLoopTest, $rulesDescription, "Avoid_FunctionCallInLoopTest");
			Erreurs::VIOLATION($FunctionCallInLoopTest__mnemo, "Function call in loop condition at line ".GetLine($condNode));
        }
        
        #~ # check all expressions using operators <, >, <=, >=, ==, !=
		my $bool_comparisonToZero;
		if ($cond !~ /\&\&|\|\|/) {
			if ( $cond =~ /([^<>!=\s]+)\s*(?:[<>!=]=|<|>)\s*([^<>!=\s]+)/ ) {
				$bool_comparisonToZero = 0;
				if (($1 =~ /(?:^|[^\.\w])0$/m) || ($2 =~ /^0(?:$|[^\.\w])/m)) {
					$bool_comparisonToZero = 1;
				}
			}
			if ((defined $bool_comparisonToZero) && ( $bool_comparisonToZero == 0)) {
				$nb_LoopCondComparedToNonZero++;
				my $numline = GetLine($condNode);
				CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_LoopCondComparedToNonZero, $rulesDescription, "LoopCondComparedToNonZero");
				Erreurs::VIOLATION($LoopCondComparedToNonZero__mnemo, "Loop condition is not comparison to zero at line ".GetLine($condNode));
			}
		}
    }
}

sub checkNestedLoop($) {
	my $loop = shift;

	# find nested loops ...
	my @subloop = GetNodesByKindList($loop, [WhileKind, ForKind, EForKind, DoKind], 1);
	for my $sloop (@subloop) {
		$nb_ToManyNestedLoop++;
		my $numline = GetLine($sloop);
		CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_ToManyNestedLoop, $rulesDescription, "Avoid_ToManyNestedLoop");
		Erreurs::VIOLATION($ToManyNestedLoop__mnemo, "sub loop at line ".GetLine($sloop));
	}
}

# HL-747 18/01/2019 C++ DIAG : Avoid overflowing loop counter 
sub CountLoop
{
    my ($fichier, $vue, $compteurs, $options) = @_;
    my $status = 0;
    $nb_OverFlowingLoopCounter = 0;
    $nb_FunctionCallInLoopTest = 0;
	$nb_LoopCondComparedToNonZero = 0;
	$nb_ToManyNestedLoop = 0;
	
    my $KindsLists = $vue->{'KindsLists'}||[];

    if ( ! defined $KindsLists )
	{
		$status |= Couples::counter_add($compteurs, $mnemo_OverFlowingLoopCounter, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $FunctionCallInLoopTest__mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $LoopCondComparedToNonZero__mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ToManyNestedLoop__mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status;
    }
    
    my $ForLoops = $KindsLists->{&ForKind};
    
	# FOR
 	for my $ForLoop (@$ForLoops) 
    {  
		my ($condNode) = @{GetChildren($ForLoop)};
        
        if ( GetKind ($condNode) eq ConditionKind )
        {          
            my $termCond = Lib::NodeUtil::GetKindData($condNode)->{'cond'};
            my $incrementCond = Lib::NodeUtil::GetKindData($condNode)->{'inc'};
            
            if ($incrementCond !~ /\+\+/ and $incrementCond !~ /\-\-/ 
                and ($termCond =~ /\=\=/ or $termCond =~ /\!\=/) )
            {
                print "+++++ Potential infinite loop risk found $incrementCond $termCond\n" if $DEBUG;
                Erreurs::VIOLATION($mnemo_OverFlowingLoopCounter, "Potential infinite loop risk found");
                $nb_OverFlowingLoopCounter++;              
            }
        }
        
        checkLoopCondition($condNode);
        checkNestedLoop($ForLoop);
    }   
    
    # WHILE
    my $WhileLoops = $KindsLists->{&WhileKind};
 	for my $WhileLoop (@$WhileLoops) {
		my ($condNode) = @{GetChildren($WhileLoop)};
		checkLoopCondition($condNode);
		checkNestedLoop($WhileLoop);
	}
	
	# DO ... WHILE
    my $DoLoops = $KindsLists->{&DoKind};
 	for my $DoLoop (@$DoLoops) {
		my $condNode = GetChildren($DoLoop)->[-1];
		checkLoopCondition($condNode);
		checkNestedLoop($DoLoop);
	}
	
	# FOR IN
    my $EForLoops = $KindsLists->{&EForKind};
 	for my $EForLoop (@$EForLoops) {
		my ($condNode) = @{GetChildren($EForLoop)};
		checkLoopCondition($condNode);
		checkNestedLoop($EForLoop);
	}

    $status |= Couples::counter_add ($compteurs, $mnemo_OverFlowingLoopCounter, $nb_OverFlowingLoopCounter);
    $status |= Couples::counter_add ($compteurs, $FunctionCallInLoopTest__mnemo, $nb_FunctionCallInLoopTest);
    $status |= Couples::dispatch ($LoopCondComparedToNonZero__mnemo, $nb_LoopCondComparedToNonZero);
    $status |= Couples::dispatch($ToManyNestedLoop__mnemo, $nb_ToManyNestedLoop);

    return $status;
}

sub CountPatternInLoop($$$) {
	my ($file, $vue, $compteurs) = @_;

	my $ret = 0;

	$nb_ConcatInLoop = 0;
	$nb_InstantiationInLoop = 0;
	$nb_OpenCloseFileInLoop = 0;

	my $root = $vue->{'structured_code'};
	my $code = \$vue->{'code'};


	if (!defined $root) {
		$ret |= Couples::counter_add($compteurs, $ConcatInLoop__mnemo, Erreurs::COMPTEUR_ERREUR_VALUE);
		$ret |= Couples::counter_add($compteurs, $InstantiationInLoop__mnemo, Erreurs::COMPTEUR_ERREUR_VALUE);
		$ret |= Couples::counter_add($compteurs, $OpenCloseFileInLoop__mnemo, Erreurs::COMPTEUR_ERREUR_VALUE);

		return $ret |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}

	my @loops = GetNodesByKindList($root, [ WhileKind, ForKind, EForKind, DoKind ], 1);
	# HL-2022 Avoid OPEN/CLOSE inside loops
	# codeBody is in lowercase
	my $regexOpen = qr/\.open\s*\(.*?\)|\bios\:\:(?:ate|binary|in|out)\b/i;
	my $regexClose = qr/\.close\b/i;
	for my $loop (@loops) {
		my $codeBody = Lib::NodeUtil::GetKindData($loop)->{'codeBody'};
		my $bodyBeginLine = Lib::NodeUtil::GetKindData($loop)->{'bodyBeginLine'};

		# check presence of literal strings concaténations (variables strings cannot be detected !)
		my $lineOffset = 0;
		while ($$codeBody =~ /(\n+)|(\bCHAINE_\d+)|(\+=)|(\+)|($regexOpen)|($regexClose)|(\bnew\b)/gc) {
			if (defined $1) {
				$lineOffset += () = $1 =~ /\n/g;
			}
			elsif (defined $7) {
				$nb_InstantiationInLoop++;
				my $numline = $bodyBeginLine + $lineOffset;
				CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_InstantiationInLoop, $rulesDescription, "Avoid_InstantiationInLoop");
				Erreurs::VIOLATION($InstantiationInLoop__mnemo, "Instantiation in loop at line " . ($bodyBeginLine + $lineOffset));
			}
			elsif (defined $2) {
				# CHAINE_XX
				if ($$codeBody =~ /\G(\s*)\+/gc) {
					$lineOffset += () = $1 =~ /\n/g;
					$nb_ConcatInLoop++;
					my $numline = $bodyBeginLine + $lineOffset;
					CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_ConcatInLoop, $rulesDescription, "Avoid_ConcatInLoop");
					Erreurs::VIOLATION($ConcatInLoop__mnemo, "String concatenation inside loop at line " . ($bodyBeginLine + $lineOffset));
				}
			}
			elsif (defined $5 || defined $6) {
				$lineOffset += () = $5 =~ /\n/g if (defined $5);
				$lineOffset += () = $6 =~ /\n/g if (defined $6);
				$nb_OpenCloseFileInLoop++;
				my $numline = $bodyBeginLine + $lineOffset;
				# print "Open/close file inside loop at line $numline\n";
				CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_OpenCloseFileInLoop, $rulesDescription, "Avoid_OpenCloseFileInLoop");
				Erreurs::VIOLATION($OpenCloseFileInLoop__mnemo, "Open/close file inside loop at line $numline");
			}
			else {
				# + first or +=
				if ($$codeBody =~ /\G(\s*)\bCHAINE_\d+(\s*)(\+)?/gc) {
					$lineOffset += () = $1 =~ /\n/g;
					$nb_ConcatInLoop++;
					my $numline = $bodyBeginLine + $lineOffset;
					CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_ConcatInLoop, $rulesDescription, "Avoid_ConcatInLoop");
					Erreurs::VIOLATION($ConcatInLoop__mnemo, "String concatenation inside loop at line " . ($bodyBeginLine + $lineOffset));

					$lineOffset += () = $2 =~ /\n/g;
					if (defined $3) {
						$nb_ConcatInLoop++;
						my $numline = $bodyBeginLine + $lineOffset;
						CommonOutput::ElaborateLog::SimpleElaborateLog($numline, Ident::Alias_ConcatInLoop, $rulesDescription, "Avoid_ConcatInLoop");
						Erreurs::VIOLATION($ConcatInLoop__mnemo, "String concatenation inside loop at line " . ($bodyBeginLine + $lineOffset));
					}
				}
			}
		}
	}

	$ret |= Couples::dispatch($ConcatInLoop__mnemo, $nb_ConcatInLoop);
	$ret |= Couples::dispatch($InstantiationInLoop__mnemo, $nb_InstantiationInLoop);
	$ret |= Couples::dispatch($OpenCloseFileInLoop__mnemo, $nb_OpenCloseFileInLoop);

	return $ret;
}

1;
