package Lib::ConditionUtil;

use strict;
use warnings;

my $LOGICAL_OR = qr/(?:\|\||\bor\b)/;
my $LOGICAL_AND = qr/(?:&&|\band\b)/;
my $COMPARISON_OPERATORS = qr/(?:<=>|<=|>=|>|<|===|==|!=|!~|=~)/;
my $LOGICAL_OPERATORS = qr/(?:$LOGICAL_OR|$LOGICAL_AND)/;
my $NEGATION_OPERATORS = qr/(?:!|\bnot\b)/;

use constant NODE_CONDITION_TYPE => 0;
use constant NODE_OPERATOR => 1;
use constant NODE_MEMBERS_LIST => 2;

use constant LEFT_MEMBER => 0;
use constant RIGHT_MEMBER => 1;

my @EqualityConditions = ();
my @IdentityConditions = ();
my @NotConditions = ();
my @OrConditions = ();
my @AndConditions = ();

my $nbAnd = 0;
my $nbOr = 0;

my $COND_TEXT;
my @ELEMENTS = ();
my $IDX = 0;

my %UNARY = (
	'not' 	=> ["not", 8],
	"!"		=> ["not", 1],
);

my %BINARY = (
	"<"		=> ["comp",		2],
	">"		=> ["comp",		2],
	"<="	=> ["comp",		2],
	">="	=> ["comp",		2],
	"=="	=> ["comp",		3],
	"==="	=> ["comp",		3],
	"<=>"	=> ["comp",		3],
	"!="	=> ["comp",		3],
	"=~"	=> ["comp",		3],
	"!~"	=> ["comp",		3],
	"&&"	=> ["logic", 	4],
	"||"	=> ["logic", 	5],
	'and'	=> ["logic",	6],
	'or'	=> ["logic",	7]
);


sub nextElement() {
	return $ELEMENTS[$IDX];
}


sub getNextElement() {
	my $element = undef;
	if ($IDX < scalar @ELEMENTS) {
		$element = $ELEMENTS[$IDX];
		$IDX++;
		while (($IDX < scalar @ELEMENTS) && ($ELEMENTS[$IDX] =~/^\s*$/m)) {
			$IDX++;
		}
	}
#print "USING : $element\n";
	return $element;
}

sub init() {
	@EqualityConditions = ();
	@IdentityConditions = ();
	@NotConditions = ();
	@OrConditions = ();
	@AndConditions = ();
	
	# init
	$IDX = 0;
	while (($IDX < scalar @ELEMENTS) && ($ELEMENTS[$IDX] =~/^\s*$/m)) {
		$IDX++;
	}
}

sub tree2String($$);
sub tree2String($$) {
	my $node = shift;
	my $indent = shift;
	
	my $output="";
	
	my $type = $node->[NODE_CONDITION_TYPE];
	my $ops  = $node->[NODE_OPERATOR];
	
	if (! defined $type) {
		$output .= "++$indent<undef>\n";
		return;
	}
	
	$output .= "++$indent<$type>";
	
	$indent .= "  ";
	if ($type eq "expr") {
		$output .= "\n";
		$output .= "++$indent$node->[1]\n";
	}
	elsif ($type eq "comp") {
		$output .= " $ops\n";
		my $members = $node->[NODE_MEMBERS_LIST];
		for my $node (@$members) {
			$output .= tree2String($node, $indent);
		}
	}
	elsif ($type eq "compList") {
		$output .= " ".join(',', @$ops)."\n";
		my $members = $node->[NODE_MEMBERS_LIST];
		for my $node (@$members) {
			$output .= tree2String($node, $indent);
		}
	}
	elsif ($type eq "not") {
		$output .= "\n";
		$output .= tree2String($node->[NODE_MEMBERS_LIST]->[0], $indent);
	}
	elsif ($type eq "logic") {
		$output .= " $ops\n";
		my $condList = $node->[NODE_MEMBERS_LIST];
		for my $node (@$condList) {
			$output .= tree2String($node, $indent);
		}
	}
	return $output;
}

sub printNode($$) {
	my $node = shift;
	my $indent = shift;
	
	print STDERR tree2String($node, $indent);
}

#-----------------------------------------------------------------------------------------------------------------

sub parseCond($);

sub parseExpressionOperand() {
#print ">> parseExpressionOperand();\n";

	my $expr = "";
	my $level = 0;
	my $next;
	while ( (defined ($next=nextElement())) && ($next !~ /^(?:$LOGICAL_OPERATORS|${COMPARISON_OPERATORS})$/m)) {
		if ($next eq '(') {
			$level++;
		}
		elsif ($next eq ')') {
			$level--;
			if ($level < 0) {
				last;
			}
		}
		$expr .= getNextElement();
	}
	my $node = ["expr", $expr];
#printNode($node, "");
	return $node;
}

sub parseParentheses() {
	getNextElement();
	my $cond = parseCond(100);
	if (nextElement() eq ')') {
		getNextElement();
	}
	else {
		print STDERR "[parseParentheses] found ".(getNextElement()||"undef")." while expecting closing parenthese!\n";
	}
	return $cond;
}

sub parseUnary() {
	my $op = getNextElement();
	my $cond = parseCond($UNARY{$op}->[1]);
	return [$UNARY{$op}->[0], $op, [$cond]];
}

# return :
# - an unary condition
# - or a binary condition
# - or an operand if none operator encountered.
sub parseCond($) {
	my $CurrentPriority = shift;

	my $next = nextElement();
	my $operand;
	
	if ($next eq '(') {
		$operand = parseParentheses();
	}	
	elsif (exists $UNARY{$next}) {
		$operand = parseUnary();
	}
	else {
		$operand = parseExpressionOperand();
	}
	
	my $op;
	
	while (defined ($op = nextElement())) {
		
		if (exists $BINARY{$op}) {
			my $priority = $BINARY{$op}->[1];
			
			if ($priority <= $CurrentPriority) {
				# consumes the operator ...
				getNextElement();
				
				# parse second operand
				my $cond = parseCond($priority);
				$operand = [$BINARY{$op}->[0], $op, [$operand, $cond]];
			}
			else {
				last;
			}
		}
		elsif ($op eq ')') {
			last;
		}
		else {
			print STDERR "[parseCond] unknow fucking operator : $op\n";
			last;
		}
	}
	
	return $operand;
}

sub parse($) {
	$COND_TEXT = shift;
#print "============= PARSING $COND_TEXT\n";
	#@ELEMENTS = split /(\s+|!=|!|&&|\|\||\b(?:not\s+in|not|and|or|in|is\s+not|is)\b|\(|\)|<=|>=|<>|>|<|==)/, $COND_TEXT;
	@ELEMENTS = split /(\s+|\(|\)|${COMPARISON_OPERATORS}|${LOGICAL_OPERATORS}|${NEGATION_OPERATORS})/, $COND_TEXT;
	
	init();
	
	my $condsTree = parseCond(100);

#printNode($condsTree, "");
	return $condsTree;
}

#-----------------------------------------------------------------------------------------------------------------

sub getNegativeImpact($);
sub getNegativeImpact($) {
	my $cond = shift;
	
	my $deltaNot = 0;
	
	if ($cond->[NODE_CONDITION_TYPE] eq "expr") {
#print STDERR "-- expr -> 1\n";
		return 1;
	}
	elsif ($cond->[NODE_CONDITION_TYPE] eq "not") {
		if ($cond->[NODE_MEMBERS_LIST]->[0]->[NODE_CONDITION_TYPE] eq "expr") {
#print STDERR "-- not expr -> - 1\n";
			return -1;
		}
		else {
			my $childDelta = getNegativeImpact($cond->[2]);
#print STDERR "-- not -> $childDelta - 1\n";
			return $childDelta - 1;
		}
	}
	elsif (($cond->[NODE_CONDITION_TYPE] eq "comp") || ($cond->[NODE_CONDITION_TYPE] eq "logic") || ($cond->[NODE_CONDITION_TYPE] eq "complist")) {
#print STDERR "-- $cond->[NODE_CONDITION_TYPE] :\n";
		for my $child (@{$cond->[NODE_MEMBERS_LIST]}) {
			my $childDelta = getNegativeImpact($child);
			#$deltaNot += getNegativeImpact($child);
			$deltaNot += $childDelta;
		}
#print STDERR "  --> total = $deltaNot\n";
		return $deltaNot;
	}	
	return 0;
}


1;
