package Ruby::CountClass;

use warnings;
use strict;

use Lib::Node;
use Lib::NodeUtil;
use Ruby::RubyNode;
use Ruby::RubyConfig;
use Ruby::CountNaming;

my $ClassImplementations_mnemo = Ident::Alias_ClassImplementations;
my $MethodsAverage_mnemo = Ident::Alias_MethodsAverage;
my $ClassVariables_mnemo = Ident::Alias_ClassVariables;
my $UtilityClassShouldBeModule_mnemo = Ident::Alias_UtilityClassShouldBeModule;
my $UnusedPrivateMethods_mnemo = Ident::Alias_UnusedPrivateMethods;
my $BadVariableNames_mnemo = Ident::Alias_BadVariableNames;
my $BadConstantNames_mnemo = Ident::Alias_BadConstantNames;
my $BadClassNames_mnemo = Ident::Alias_BadClassNames;
my $VariableNameLengthAverage_mnemo = Ident::Alias_VariableNameLengthAverage;
my $TotalAttributes_mnemo = Ident::Alias_TotalAttributes;
my $ShortVarName_mnemo = Ident::Alias_ShortVarName;
my $ShortClassNamesLT_mnemo = Ident::Alias_ShortClassNamesLT;
my $ClassNameLengthAverage_mnemo = Ident::Alias_ClassNameLengthAverage;
my $LocalVariables_mnemo = Ident::Alias_LocalVariables;
my $ConstantDefinitions_mnemo = Ident::Alias_ConstantDefinitions;

my $nb_ClassImplementations = 0;
my $nb_MethodsAverage = 0;
my $nb_ClassVariables = 0;
my $nb_UtilityClassShouldBeModule = 0;
my $nb_UnusedPrivateMethods = 0;
my $nb_BadVariableNames = 0;
my $nb_BadConstantNames = 0;
my $nb_BadClassNames = 0;
my $nb_VariableNameLengthAverage = 0;
my $nb_TotalAttributes = 0;
my $nb_ShortVarName = 0;
my $nb_ShortClassNamesLT = 0;
my $nb_ClassNameLengthAverage = 0;
my $nb_LocalVariables = 0;
my $nb_ConstantDefinitions = 0;

sub checkNamespace($$);
sub checkNamespace($$) {
	my $struct = shift;
	my $namespaces = shift;
	
	my @artifacts = GetNodesByKindList_StopAtBlockingNode($struct, [ClassKind, ModuleKind, MethodKind], [ClassKind, ModuleKind]);
	
	my $methods = [];
	my @structs = ();
	for my $artifact (@artifacts) {
		if (IsKind($artifact, MethodKind)) {
			push @$methods, $artifact;
		}
		else {
			$namespaces->{GetName($artifact)} = $artifact;
			push @structs, $artifact;
		}
	}
	for my $method (@$methods) {
		my $name = GetName($method);
#print STDERR "CHECK method $name with namespaces ".(join ",", keys %$namespaces)."\n";
		if ($name =~ /\A(\w+)::/) {
			if (exists $namespaces->{$1}) {
				print STDERR "Method $name at line ".GetLine($method)." is re-openning namespace $1 declared at line ".GetLine($namespaces->{$1})."\n";
			}
		}
	}
	
	for my $struct (@structs) {
		my $name = GetName($struct);
#print STDERR "CHECK method $name with namespaces ".(join ",", keys %$namespaces)."\n";
		if ($name =~ /\A(\w+)::/) {
			if (exists $namespaces->{$1}) {
				print STDERR "class/module $name at line ".GetLine($struct)." is re-openning namespace $1 declared at line ".GetLine($namespaces->{$1})."\n";
			}
		}
	}
	
	for my $struct (@structs) {
		checkNamespace($struct, $namespaces);
	}
}

sub checkClassOrInstanceVariable($$$$) {
	my $artifactBuffer = shift;
	my $H_vars = shift;
	my $line = shift;
	my $structKind = shift;

	while ( $$artifactBuffer =~ /(\n)|\@(\@)?(\w+)/g) {
		if (defined $1) {
			$line++;
		}
		elsif (defined $2) {
			if (! exists $H_vars->{"class"}->{$3}) {
				my $varName = $3;
				$H_vars->{"class"}->{$varName} = $line;
				
				$nb_VariableNameLengthAverage += length($varName);
				
				if (length($varName) < Ruby::RubyConfig::MIN_VARIABLE_NAME_LENGTH) {
					$nb_ShortVarName++;
					Erreurs::VIOLATION($ShortVarName_mnemo, "Short class variable name $varName at line $line");
				}
				
				Erreurs::VIOLATION($ClassVariables_mnemo, "Class variable $varName at line $line");
				$nb_ClassVariables++;
				if (! Ruby::CountNaming::isSnakecase($3)) {
					$nb_BadVariableNames++;
					Erreurs::VIOLATION($BadVariableNames_mnemo, "Bad class variable name $varName at line $line");
				}
			}
		}
		else {
			if (! exists $H_vars->{"instance"}->{$3}) {
				$H_vars->{"instance"}->{$3} = $line;
				
				$nb_VariableNameLengthAverage += length($3);
				
				if (length($3) < Ruby::RubyConfig::MIN_VARIABLE_NAME_LENGTH) {
					$nb_ShortVarName++;
					Erreurs::VIOLATION($ShortVarName_mnemo, "Short instance variable name $3 at line $line");
				}
				
				$nb_TotalAttributes++;
				
#print STDERR "DEBUG : INSTANCE variable \@$3 at line $line\n";
				if (! Ruby::CountNaming::isSnakecase($3)) {
					$nb_BadVariableNames++;
					Erreurs::VIOLATION($BadVariableNames_mnemo, "Bad instance variable name $3 at line $line");
				}
			}
		}
	}
}


# capture also global, class and instances var, becasue they may be mixed with local and constant inside a dereferncement list !!!
my $REG_IDENTIFIERS_LIST = qr/(?:[\*\$\@]*\w+(?:\s*,\s*)?)+/;

sub checkConstantOrLocalVariable($$$;$) {
	my $artifactBuffer = shift;
	my $H_vars = shift;
	my $line = shift;
	my $parameters = shift; #if method

	my $artifactLine = $line;
	my @localVar = ();
	
	# search identifiers initialized with =
	# Operator +=, -=, ... are considered too, because only a variable can be affected by these operators.
	# FIXME : known limitations are cascad initializations like : a = b = 0 (=> here b will not be found)
	while ( $$artifactBuffer =~ /(?:\A|;|=|\(|(\n)|\b(?:if|elsif|then|while|and|or|unless|case|when)\b|&&|\|\||:[ \t]+)[ \t]*($REG_IDENTIFIERS_LIST)\s*(?:([\+\*\-\/\%\&\|]=)|=[^=~])|(\n)/g) {
		if (defined $4) {
			$line++;
		}
		else {
			if (defined $1) {
				$line++;
			}

			my  @var_list = split /\s*,\s*/, $2;
			
			for my $varName (@var_list) {
				
				# do not consider class and global var (captured for conveniance, but useless) ...
				# do not consider fake variables used for masking literal strings
				next if ($varName =~ /\A[\@\$]|CHAINE_\d+/);
				
				# remove "several element modifier" ...
				$varName =~ s/\A\*//;
				
				if ($varName =~ /\A[A-Z]/) {
					if (! exists $H_vars->{"constant"}->{$varName}) {
						$nb_ConstantDefinitions++;
						$H_vars->{"constant"}->{$varName} = $line;
						
						$nb_VariableNameLengthAverage += length($varName);
						
						if (length($varName) < Ruby::RubyConfig::MIN_VARIABLE_NAME_LENGTH) {
							$nb_ShortVarName++;
							Erreurs::VIOLATION($ShortVarName_mnemo, "Short constant variable name $varName at line $line");
						}
						
#print STDERR "DEBUG : CONSTANT $varName at line $line\n";
						if (! Ruby::CountNaming::isUppercase($varName)) {
							$nb_BadConstantNames++;
							Erreurs::VIOLATION($BadConstantNames_mnemo, "Bad constant name $varName at line $line");
						}
					}
				}
				else {
					if (! exists $H_vars->{"local"}->{$varName}) {
						if ((!defined $parameters) || (! exists $parameters->{$varName})) {
							
							$nb_LocalVariables++;
							$H_vars->{"local"}->{$varName} = $line;
							
							$nb_VariableNameLengthAverage += length($varName);
						
							if (length($varName) < Ruby::RubyConfig::MIN_VARIABLE_NAME_LENGTH) {
								$nb_ShortVarName++;
								Erreurs::VIOLATION($ShortVarName_mnemo, "Short local variable name $varName at line $line");
							}
							
#print STDERR "DEBUG : LOCAL VAR $varName at line $line\n";
							if (! Ruby::CountNaming::isSnakecase($varName)) {
								$nb_BadVariableNames++;
								Erreurs::VIOLATION($BadVariableNames_mnemo, "Bad local variable name $varName at line $line");
							}
						}
					}
				}
			}
		}
	}

    # FIXME : abandonned, because too many false positive, due (at least) to variables initialized with destructuring assignment.
	#~ for my $var (keys %{$H_vars->{"local"}}) {
		
		#~ next if ($var eq "_");
		
		#~ if ($$artifactBuffer !~ /(\b$var\s*=[=~>])|(\b$var\b\s*(?:[^\s=]|\z))/) {
			#~ print STDERR "DEBUG : unused variable $var declared at line ".$H_vars->{"local"}->{$var}."\n";
		#~ }
	#~ }


	# WAS USED FOR DEBUG (show potential forgoten variables by using another algorithm ...)
	#~ $line = $artifactLine;
	#~ while ( $$artifactBuffer =~ /(?:\A|(\n)|(\bdo +\|)|[^\@\$\.:])[ \t]*\b((?:\w+(?:\s*,\s*)?)+)\s*(?:[\+\*\-\/\%\&\|]=|=[^=~>])|(\n)/g) {
		
		#~ next if defined $2;
		
		#~ if (defined $4) {
			#~ $line++;
		#~ }
		#~ else {
			#~ if (defined $1) {
				#~ $line++;
			#~ }
			#~ my $varName = $3;
			#~ next if ($varName =~ /,/);
			#~ if ((! exists $H_vars->{"constant"}->{$varName}) && (! exists $H_vars->{"local"}->{$varName}) && ((!defined $parameters) || (! exists $parameters->{$varName}))) {
#~ print STDERR "DEBUG : MISSING VAR $varName at line $line ???\n";
			
			#~ }
		#~ }
	#~ }
}

sub CountClass($$$) {
	my ($fichier, $views, $compteurs, $options) = @_;
	
	my $status = 0;

	$nb_ClassImplementations = 0;
	$nb_MethodsAverage = 0;
	$nb_ClassVariables = 0;
	$nb_UtilityClassShouldBeModule = 0;
	$nb_UnusedPrivateMethods = 0;
	$nb_BadVariableNames = 0;
	$nb_BadConstantNames = 0;
	$nb_BadClassNames = 0;
	$nb_VariableNameLengthAverage = 0;
	$nb_TotalAttributes = 0;
	$nb_ShortVarName = 0;
	$nb_ShortClassNamesLT = 0;
	$nb_ClassNameLengthAverage = 0;
	$nb_LocalVariables = 0;
	$nb_ConstantDefinitions = 0;
	
	my $root = $views->{'structured_code'};
	my $KindsLists = $views->{'KindsLists'};
	my $agglo_index = $views->{'agglo_LinesIndex'};
	my $agglo = \$views->{'agglo'};
	my $artifact = $views->{'artifact'};
	my $HString = $views->{'HString'};

	if ( ! defined $KindsLists ) {
		$status |= Couples::counter_add($compteurs, $ClassImplementations_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $MethodsAverage_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ClassVariables_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $UtilityClassShouldBeModule_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $UnusedPrivateMethods_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $BadVariableNames_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $BadConstantNames_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $BadClassNames_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $VariableNameLengthAverage_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $TotalAttributes_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ShortVarName_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ShortClassNamesLT_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ClassNameLengthAverage_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $LocalVariables_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ConstantDefinitions_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	my $classes = $KindsLists->{&ClassKind};
	my $modules = $KindsLists->{&ModuleKind};
	
	my @classModule = (@$classes, @$modules);
	
	my %H_vars = ("class" => {}, "instance" => {}, "constant" => {}, "local" => {});
	
	
	# CHECK constant & local variables in ROOT & FUNCTIONS		
	checkConstantOrLocalVariable(\$artifact->{"root_1"}, \%H_vars, GetLine($root));
	my @functions = GetNodesByKindList_StopAtBlockingNode($root, [MethodKind], [MethodKind, ClassKind, ModuleKind]);
	$H_vars{"local"} = {};
	for my $func (@functions) {
		checkConstantOrLocalVariable(
						Ruby::RubyNode::getRubyKindData($func, 'codeWithStrings'),
						\%H_vars,
						GetLine($func),
						Ruby::RubyNode::getRubyKindData($func, 'H_parameters'));
	}
	
	# CHECK classes & modules
	for my $struct ( @classModule) {
		
		my $structKind = GetKind($struct);
		
		$nb_ClassImplementations++;
		
		my $className = GetName($struct);
		my $statement = GetStatement($struct);

		# CHECK NAME ...
		my $nameToCheck = $className;
		# remove namespace part ...
		$nameToCheck =~ s/\A.*:://;
		if ($nameToCheck !~ /_singleton_class_line_/) {
			if (! Ruby::CountNaming::isPascalcase($nameToCheck)) {
				$nb_BadClassNames++;
				Erreurs::VIOLATION($BadClassNames_mnemo, "bad class name $className at line ".GetLine($struct));
			}
		}
		
		$nb_ClassNameLengthAverage += length($nameToCheck);
		
		if (length($nameToCheck) < Ruby::RubyConfig::MIN_CLASS_NAME_LENGTH) {
			$nb_ShortClassNamesLT++;
			Erreurs::VIOLATION($ShortClassNamesLT_mnemo, "Short class name $nameToCheck at line ".GetLine($struct));
		}
		
		my $artikey = Ruby::RubyNode::getRubyKindData($struct, 'artifact_key');
		#~ my $codeWithString = $artifact->{$artikey};
		#~ $codeWithString =~ s/(\bCHAINE_\d+\b)/$HString->{$1}/gx;

		my @methods = GetNodesByKindList_StopAtBlockingNode($struct, [MethodKind], [MethodKind, ClassKind, ModuleKind]);
		
		my $is_inherited = 0;
		if ($$statement =~ /</) {
			$is_inherited = 1;
		}
		
		
		my $nb_methods=0;
		my $has_instance_methods = 0;
		my $has_initialize_methods = 0;
		
		my @privateMethods = ();
		
		# get infos about methods
		for my $method (@methods) {
			my $methodName = GetName($method);
			
			if ($methodName eq "initialize") {
				$has_initialize_methods = 1;
				next;
			}
			
			if ((!$has_instance_methods) && !(($methodName =~ /\Aself\./) || (defined Ruby::RubyNode::getRubyKindData($method, "self_singleton_method")))) {
				$has_instance_methods = 1;	
			}
			
			my $visibility = Ruby::RubyNode::getRubyKindData($method, 'visibility');
			if ((defined $visibility) && ($visibility eq "private")) {
				push @privateMethods, $method;
			}
			
			$nb_methods++;
		}
		
		# CHECK PRIVATE METHODS USAGE
		for my $privateMeth (@privateMethods) {
			my $pattern = quotemeta GetName($privateMeth);
			$pattern =~ s/\A(\w)/\\b$1/;
			$pattern =~ s/(\w)$/$1\\b/m;

			# search private method in each method of the class (except itself) !
			my $used = 0;
			for my $method (@methods) {
				next if ($method == $privateMeth);

				my $codeWithStrings = Ruby::RubyNode::getRubyKindData($method, 'codeWithStrings');

				while ($$codeWithStrings =~ /(?:\A|:|\bself\.|[^\.\@])$pattern/) {
					$used = 1;
					last;
				}
			}
			if (!$used) {
				$nb_UnusedPrivateMethods++;
				Erreurs::VIOLATION($UnusedPrivateMethods_mnemo, "Unused private method : ".GetName($privateMeth)." at line ".GetLine($privateMeth));
			}
		}
		
		$nb_MethodsAverage += $nb_methods;
		
		#~ if ($nb_methods > Ruby::RubyConfig::MAX_METHODS) {
			#~ $nb_ClassVariables++;
			#~ Erreurs::VIOLATION($ClassVariables_mnemo, "Complex class or module ($className) with $nb_methods methods at line ".GetLine($struct));
		#~ }
		
		
		
		
		# CHECK VARIABLES : class, instance, constant, local.
		
		# search in class or module (module can be inside a class, and even useless can have their own class variable)...
#print STDERR "CHECK $artikey\n";
		$H_vars{"class"} = {};
		$H_vars{"instance"} = {};
		$H_vars{"local"} = {};
		checkClassOrInstanceVariable(\$artifact->{$artikey}, \%H_vars, GetLine($struct), $structKind);
		checkConstantOrLocalVariable(\$artifact->{$artikey}, \%H_vars, GetLine($struct));
		# search in each method ... 
		for my $method (@methods) {
#print STDERR "CHECK method".GetKind($method)."\n";
			$H_vars{"local"} = {};
			checkClassOrInstanceVariable(Ruby::RubyNode::getRubyKindData($method, 'codeWithStrings'), \%H_vars, GetLine($method), $structKind);
			checkConstantOrLocalVariable(
						Ruby::RubyNode::getRubyKindData($method, 'codeWithStrings'),
						\%H_vars,
						GetLine($method),
						Ruby::RubyNode::getRubyKindData($method, 'H_parameters'));
		}
		
			
		if ($structKind eq ClassKind) {	
			# UTLITY CHECKING
			if (($className !~ /::/) && (!$is_inherited) && ($nb_methods) && (!$has_instance_methods) && (!$has_initialize_methods)) {
				$nb_UtilityClassShouldBeModule++;
				Erreurs::VIOLATION($UtilityClassShouldBeModule_mnemo, "Utility class should be module at line ".GetLine($struct)."\n");
			}
		}
	}
	
	# For robustness (symptomatic of a parsing error) and information, check class variable outside class !!
	my @funcs = GetNodesByKindList_StopAtBlockingNode($root, [MethodKind], [MethodKind, ClassKind, ModuleKind]);
	for my $func (@funcs) {
		my $body = Ruby::RubyNode::getRubyKindData($func, 'codeBody');
		if ($$body =~ /\@\@(\w+)/) {
print STDERR "DEBUG : CLASS VARIABLE inside a method ".GetName($func)." external to a class !!!\n";
		}
	}
	
	checkNamespace($root, {});
	
	
	if ($nb_ClassImplementations) {
		$nb_MethodsAverage 			= int($nb_MethodsAverage		/	$nb_ClassImplementations);
		$nb_ClassNameLengthAverage 	= int($nb_ClassNameLengthAverage/	$nb_ClassImplementations);
	}

	my $nb_TotalVar = $nb_LocalVariables + $nb_ConstantDefinitions + $nb_ClassVariables + $nb_TotalAttributes;
	if ($nb_TotalVar) {
		$nb_VariableNameLengthAverage = int($nb_VariableNameLengthAverage / $nb_TotalVar);
	}

	#~ my $nb_BadVariableNameAverage=0;
	#~ my $nb_BadConstantNameAverage=0;
	#~ if ($nb_LocalVariables) {
		#~ $nb_BadVariableNameAverage = int(($nb_BadVariableNames*100)/$nb_LocalVariables);
	#~ }
	#~ if ($nb_ConstantDefinitions) {
		#~ $nb_BadConstantNameAverage = int(($nb_BadConstantNames*100)/$nb_ConstantDefinitions);
	#~ }
	#~ print STDERR "Bad constant name average = $nb_BadConstantNameAverage%\n";
	#~ print STDERR "Bad variable name average = $nb_BadVariableNameAverage%\n";
	
	Erreurs::VIOLATION($ClassImplementations_mnemo, "METRIC : nb classes : $nb_ClassImplementations");
	Erreurs::VIOLATION($MethodsAverage_mnemo, "METRIC : class/module methods average : $nb_MethodsAverage");
		
	$status |= Couples::counter_add($compteurs, $ClassImplementations_mnemo, $nb_ClassImplementations);
	$status |= Couples::counter_add($compteurs, $MethodsAverage_mnemo, $nb_MethodsAverage);
	$status |= Couples::counter_add($compteurs, $ClassVariables_mnemo, $nb_ClassVariables);
	$status |= Couples::counter_add($compteurs, $UtilityClassShouldBeModule_mnemo, $nb_UtilityClassShouldBeModule);
	$status |= Couples::counter_add($compteurs, $UnusedPrivateMethods_mnemo, $nb_UnusedPrivateMethods);
	$status |= Couples::counter_add($compteurs, $BadVariableNames_mnemo, $nb_BadVariableNames);
	$status |= Couples::counter_add($compteurs, $BadConstantNames_mnemo, $nb_BadConstantNames);
	$status |= Couples::counter_add($compteurs, $BadClassNames_mnemo, $nb_BadClassNames);
	$status |= Couples::counter_add($compteurs, $VariableNameLengthAverage_mnemo, $nb_VariableNameLengthAverage);
	$status |= Couples::counter_add($compteurs, $TotalAttributes_mnemo, $nb_TotalAttributes);
	$status |= Couples::counter_add($compteurs, $ShortVarName_mnemo, $nb_ShortVarName);
	$status |= Couples::counter_add($compteurs, $ShortClassNamesLT_mnemo, $nb_ShortClassNamesLT);
	$status |= Couples::counter_add($compteurs, $ClassNameLengthAverage_mnemo, $nb_ClassNameLengthAverage);
	$status |= Couples::counter_add($compteurs, $LocalVariables_mnemo, $nb_LocalVariables);
	$status |= Couples::counter_add($compteurs, $ConstantDefinitions_mnemo, $nb_ConstantDefinitions);
	
	return $status;
}

1;


