package Ruby::CountMethod;

use warnings;
use strict;

use Lib::Node;
use Lib::NodeUtil;
use Ruby::RubyNode;
use Ruby::RubyConfig;
use Ruby::CountNaming;


use constant PARAM_NAME => 0;
use constant PARAM_INIT => 1;
use constant PARAM_TYPE => 2;
use constant PARAM_MODE => 3;

my $NestedRoutines_mnemo = Ident::Alias_NestedRoutines;
my $MethodImplementations_mnemo = Ident::Alias_MethodImplementations;
my $MethodsLengthAverage_mnemo = Ident::Alias_MethodsLengthAverage;
my $EmptyMethods_mnemo = Ident::Alias_EmptyMethods;
my $LongRoutines_mnemo = Ident::Alias_LongRoutines;
my $OptionalParamsBeforeRequired_mnemo = Ident::Alias_OptionalParamsBeforeRequired;
my $TopLevelRoutines_mnemo = Ident::Alias_TopLevelRoutines;
my $MultiLinesExpressionRoutines_mnemo = Ident::Alias_MultiLinesExpressionRoutines;
my $MissingParenthesesForParams_mnemo = Ident::Alias_MissingParenthesesForParams;
my $UselessParenthesesForParams_mnemo = Ident::Alias_UselessParenthesesForParams;
my $MissingLambdaLiteral_mnemo = Ident::Alias_MissingLambdaLiteral;
my $WithTooMuchParametersMethods_mnemo = Ident::Alias_WithTooMuchParametersMethods;
my $ParametersAverage_mnemo = Ident::Alias_ParametersAverage;
my $UnusedParameters_mnemo = Ident::Alias_UnusedParameters;
my $BadMethodNames_mnemo = Ident::Alias_BadMethodNames;
my $ShortMethodNamesLT_mnemo = Ident::Alias_ShortMethodNamesLT;
my $MethodNameLengthAverage_mnemo = Ident::Alias_MethodNameLengthAverage;

my $nb_NestedRoutines = 0;
my $nb_MethodImplementations = 0;
my $nb_MethodsLengthAverage = 0;
my $nb_EmptyMethods = 0;
my $nb_LongRoutines = 0;
my $nb_OptionalParamsBeforeRequired = 0;
my $nb_TopLevelRoutines = 0;
my $nb_MultiLinesExpressionRoutines = 0;
my $nb_MissingParenthesesForParams = 0;
my $nb_UselessParenthesesForParams = 0;
my $nb_MissingLambdaLiteral = 0;
my $nb_WithTooMuchParametersMethods = 0;
my $nb_ParametersAverage = 0;
my $nb_UnusedParameters = 0;
my $nb_BadMethodNames = 0;
my $nb_ShortMethodNamesLT = 0;
my $nb_MethodNameLengthAverage = 0;

sub checkMethodLength($) {
	my $method = shift;
	my $codeBody = Ruby::RubyNode::getRubyKindData($method, 'codeBody');
			
	my $nb_LOC = () = $$codeBody =~ /^.*\S/mg;
	if ($nb_LOC > 0) {
		if ($nb_LOC > Ruby::RubyConfig::MAX_METHOD_LENGTH) {
			Erreurs::VIOLATION($LongRoutines_mnemo, "Long method ($nb_LOC lines) at line ".GetLine($method));
			$nb_LongRoutines++;
		}

		$nb_MethodsLengthAverage += $nb_LOC;
		
		# Endless methods ...
		if (${GetStatement($method)} =~ /=\s*\z/) {
			if ($nb_LOC > 1) {
				$nb_MultiLinesExpressionRoutines ++;
				Erreurs::VIOLATION($MultiLinesExpressionRoutines_mnemo, "Endless method with several line at line ".GetLine($method));
			}
		}
	}
	else {
		$nb_EmptyMethods ++;
		Erreurs::VIOLATION($EmptyMethods_mnemo, "Empty method at line ".GetLine($method));
	}
}

sub checkParameters($$) {
	my $method = shift;
	my $HString = shift;
	
	my $params = Ruby::RubyNode::getRubyKindData($method, 'parameters');
	my $proto = GetStatement($method);

	if (scalar @$params) {
		if ( $$proto !~ /\)\s*/m) {
			$nb_MissingParenthesesForParams++;
			Erreurs::VIOLATION($MissingParenthesesForParams_mnemo, "Missing parentheses for method parameters at line ".GetLine($method));
		}
	}
	elsif ( $$proto =~ /\)\s*/m) {
		$nb_UselessParenthesesForParams++;
		Erreurs::VIOLATION($UselessParenthesesForParams_mnemo, "Unexpected parentheses for method without parameters at line ".GetLine($method));
	}
	
	my $body = Ruby::RubyNode::getRubyKindData($method, 'codeBody');
	
	my $args = Ruby::RubyNode::getRubyKindData($method, "parameters");
	my $opt_flag = 0;
	for my $arg (@$args) {
		my $type = $arg->[PARAM_TYPE];
		my $name = $arg->[PARAM_NAME];
		my $init = $arg->[PARAM_INIT];
		my $mode = $arg->[PARAM_MODE];
			
		if (($init ne "") || ($type eq "*") || ($type eq "**")) {
			$opt_flag = 1;
		}
		# NOTE : block are always in last position. So make an exception for prefix "&"
		elsif (($opt_flag) && ($type ne "&")){
			if (defined $name) {
				Erreurs::VIOLATION($OptionalParamsBeforeRequired_mnemo, "Required parameter ($name) after optional parameter at line ".GetLine($method));
			}
			else {
				Erreurs::VIOLATION($OptionalParamsBeforeRequired_mnemo, "Required anonymous parameter after optional parameter at line ".GetLine($method));
			}
			$nb_OptionalParamsBeforeRequired++;
		}
#print STDERR "PARAMETER $name\n";
		if (($name !~ /\A(_|\@|\$)/) && ($type ne "&")) {
			if (defined $body) {
#print STDERR "BODY = $$body\n";
				if ($$body !~ /[^\.:]\b$name\b[^:\?]/) {
				#if ($$body !~ /\b$name\b/) {
					my $found_in_strings = 0;
					pos($$body) = undef;
					while ($$body =~ /\b(CHAINE_\d+)\b/g) {
#rint STDERR "STRING  = $HString->{$1}, SEARCHING $name\n";
						if ($HString->{$1} =~ /#\{$name\}/) {
							$found_in_strings = 1;
							last;
						}
					}
					if (!$found_in_strings) {
						$nb_UnusedParameters++;
						Erreurs::VIOLATION($UnusedParameters_mnemo, "Unsused parameter '$name' at line ".GetLine($method));
					}
				}
			}
		}
	}	
}

sub CountMethod($$$) {
	my ($fichier, $views, $compteurs, $options) = @_;
	
	my $status = 0;

	$nb_NestedRoutines = 0;
	$nb_MethodImplementations = 0;
	$nb_MethodsLengthAverage = 0;
	$nb_EmptyMethods = 0;
	$nb_LongRoutines = 0;
	$nb_OptionalParamsBeforeRequired = 0;
	$nb_TopLevelRoutines = 0;
	$nb_MultiLinesExpressionRoutines = 0;
	$nb_MissingParenthesesForParams = 0;
	$nb_UselessParenthesesForParams = 0;
	$nb_WithTooMuchParametersMethods = 0;
	$nb_ParametersAverage = 0;
	$nb_UnusedParameters = 0;
	$nb_BadMethodNames = 0;
	$nb_ShortMethodNamesLT = 0;
	$nb_MethodNameLengthAverage = 0;
	
	my $KindsLists = $views->{'KindsLists'};
	my $HString = $views->{'HString'};

	if ( ! defined $KindsLists ) {
		$status |= Couples::counter_add($compteurs, $NestedRoutines_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $MethodImplementations_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $MethodsLengthAverage_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $EmptyMethods_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $LongRoutines_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $OptionalParamsBeforeRequired_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $TopLevelRoutines_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $MultiLinesExpressionRoutines_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $MissingParenthesesForParams_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $UselessParenthesesForParams_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $WithTooMuchParametersMethods_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ParametersAverage_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $UnusedParameters_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $BadMethodNames_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $ShortMethodNamesLT_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $MethodNameLengthAverage_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	my $Methods = $KindsLists->{&MethodKind};
	
	for my $method (@$Methods) {
		
		my $name = GetName($method);
		
		next if $name eq "initialize";
		
		$nb_MethodImplementations++;
		
		my $methodLine = GetLine($method);
		my $statement = GetStatement($method);
#print STDERR "PROTO : $$statement\n";
		
		my $nameToCheck = $name;
		
		$nb_MethodNameLengthAverage += length($nameToCheck);
		
		# keep only the last part of the name ...
		$nameToCheck =~ s/.*\.//;
		if (! Ruby::CountNaming::isSnakecase($nameToCheck)) {
			$nb_BadMethodNames++;
			Erreurs::VIOLATION($BadMethodNames_mnemo, "Bad method name '$nameToCheck' at line $methodLine");
		}
		
				
		if (length($nameToCheck) < Ruby::RubyConfig::MIN_METHOD_NAME_LENGTH) {
			$nb_ShortMethodNamesLT++;
			Erreurs::VIOLATION($ShortMethodNamesLT_mnemo, "Short method name $nameToCheck at line $methodLine");
		}
		
		# search NESTED METHODS
		my @nestedList = GetNodesByKindList_StopAtBlockingNode($method, [MethodKind], [ClassKind, ModuleKind, MethodKind]);
		for my $nested (@nestedList) {
			$nb_NestedRoutines++;
			Erreurs::VIOLATION($NestedRoutines_mnemo, "Nested method at line ".GetLine($nested));
		}
		
		checkMethodLength($method);
		
		if (IsKind(GetParent($method), RootKind)) {
			$nb_TopLevelRoutines++;
			Erreurs::VIOLATION($TopLevelRoutines_mnemo, "Top level method at line ".GetLine($method));
		}
		
		# PARAMETERS
		checkParameters($method, $HString);
		
		my $nb_params = scalar @{Ruby::RubyNode::getRubyKindData($method, 'parameters')};
		$nb_ParametersAverage += $nb_params;
		
		if ($nb_params> Ruby::RubyConfig::MAX_PARAMETERS) {
			$nb_WithTooMuchParametersMethods++;
			Erreurs::VIOLATION($WithTooMuchParametersMethods_mnemo, "To many parameters for method $name at line ".GetLine($method));
		}
		
		# ----------
		my $parent = GetParent($method);
		my $kindParent = GetKind($parent);
		my $nameParent = GetName($parent);
		if (($kindParent eq ModuleKind) || ($kindParent eq ClassKind)) {
			my $regexp = quotemeta $nameParent;
			if ($name =~ /\A${regexp}::/) {
				print STDERR "Missing self prefix\n";
			}
		}
	}

	if ($nb_MethodImplementations > 0) {
		$nb_ParametersAverage 		= int($nb_ParametersAverage 		/ $nb_MethodImplementations);
		$nb_MethodNameLengthAverage = int($nb_MethodNameLengthAverage 	/ $nb_MethodImplementations);
	}
	
	if ( ($nb_MethodImplementations - $nb_EmptyMethods) > 0) {
		$nb_MethodsLengthAverage = int(($nb_MethodsLengthAverage) / ($nb_MethodImplementations - $nb_EmptyMethods));
	}
	else {
		$nb_MethodsLengthAverage = 0;
	}

	my $nb_BadMethodNameAverage = 0;
	if ($nb_MethodImplementations) {
		$nb_BadMethodNameAverage = int($nb_BadMethodNames / $nb_MethodImplementations);
	}

	Erreurs::VIOLATION($MethodsLengthAverage_mnemo, "METRIC : methods length average : $nb_MethodsLengthAverage");
	Erreurs::VIOLATION($ParametersAverage_mnemo, "METRIC : methods parameters average : $nb_ParametersAverage");
	
	$status |= Couples::counter_add($compteurs, $NestedRoutines_mnemo, $nb_NestedRoutines);
	$status |= Couples::counter_add($compteurs, $MethodImplementations_mnemo, $nb_MethodImplementations);
	$status |= Couples::counter_add($compteurs, $MethodsLengthAverage_mnemo, $nb_MethodsLengthAverage);
	$status |= Couples::counter_add($compteurs, $EmptyMethods_mnemo, $nb_EmptyMethods);
	$status |= Couples::counter_add($compteurs, $LongRoutines_mnemo, $nb_LongRoutines);
	$status |= Couples::counter_add($compteurs, $OptionalParamsBeforeRequired_mnemo, $nb_OptionalParamsBeforeRequired);
	$status |= Couples::counter_add($compteurs, $TopLevelRoutines_mnemo, $nb_TopLevelRoutines);
	$status |= Couples::counter_add($compteurs, $MultiLinesExpressionRoutines_mnemo, $nb_MultiLinesExpressionRoutines);
	$status |= Couples::counter_add($compteurs, $WithTooMuchParametersMethods_mnemo, $nb_WithTooMuchParametersMethods);
	$status |= Couples::counter_add($compteurs, $ParametersAverage_mnemo, $nb_ParametersAverage);
	$status |= Couples::counter_add($compteurs, $UnusedParameters_mnemo, $nb_UnusedParameters);
	$status |= Couples::counter_add($compteurs, $BadMethodNames_mnemo, $nb_BadMethodNames);
	$status |= Couples::counter_add($compteurs, $ShortMethodNamesLT_mnemo, $nb_ShortMethodNamesLT);
	$status |= Couples::counter_add($compteurs, $MethodNameLengthAverage_mnemo, $nb_MethodNameLengthAverage);

	# countLambda id called here because of common mnemonic with CountMethod...
	$status |= countLambda($fichier, $views, $compteurs, $options);
	
	$status |= Couples::counter_add($compteurs, $MissingParenthesesForParams_mnemo, $nb_MissingParenthesesForParams);
	$status |= Couples::counter_add($compteurs, $UselessParenthesesForParams_mnemo, $nb_UselessParenthesesForParams);
	
	return $status;
}

sub countLambda($$$) {
	my ($fichier, $views, $compteurs, $options) = @_;
	
	my $status = 0;
	
	$nb_MissingLambdaLiteral = 0;
	
	my $KindsLists = $views->{'KindsLists'};

	if ( ! defined $KindsLists ) {
		$status |= Couples::counter_add($compteurs, $MissingLambdaLiteral_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	my $Lambdas = $KindsLists->{&LambdaKind};
	
	for my $lambda (@$Lambdas) {

		my $statement = GetStatement($lambda);
		my $name = GetName($lambda);
		
		my $flag_literalSyntax = 0;
		if ($$statement =~ /\A->/) {
			$flag_literalSyntax = 1;
		}
		
		# BODY
		my $codeBody = Ruby::RubyNode::getRubyKindData($lambda, 'codeBody');
		my $nb_LOC = () = $$codeBody =~ /^.*\S/mg;

		if (($nb_LOC == 1) && ($flag_literalSyntax == 0)) {
			Erreurs::VIOLATION($MissingLambdaLiteral_mnemo, "Missing literal syntax for single line lambda at line ".GetLine($lambda));
			$nb_MissingLambdaLiteral++;
		}
		
		# PARAMETERS
		if ($flag_literalSyntax == 1) {
			my $params = Ruby::RubyNode::getRubyKindData($lambda, 'parameters');

			if ($$statement =~ /\A->\s*\(/) {
				if (scalar @$params == 0) {
					Erreurs::VIOLATION($UselessParenthesesForParams_mnemo, "Useless parentheses for lambda without parameters at line ".GetLine($lambda));
					$nb_UselessParenthesesForParams++;
				}
			}
			else {
				if (scalar @$params > 0) {
					Erreurs::VIOLATION($MissingParenthesesForParams_mnemo, "Missing parentheses for lambda parameters at line ".GetLine($lambda));
					$nb_MissingParenthesesForParams++;
				}
			}
		}
	}
		
	$status |= Couples::counter_add($compteurs, $MissingLambdaLiteral_mnemo, $nb_MissingLambdaLiteral);
	
	return $status;
}
1;


