package Ruby::CountRuby;

use strict;
use warnings;

use Erreurs;
use Lib::NodeUtil;
use Ruby::RubyNode;

my $BadLogicIndents_mnemo = Ident::Alias_BadLogicIndents();
my $TabInsideIndentation_mnemo = Ident::Alias_TabInsideIndentation();
my $BadIndentationStyle_mnemo = Ident::Alias_BadIndentationStyle();
my $UnexpectedSemicolon_mnemo = Ident::Alias_UnexpectedSemicolon();
my $MultipleStatementsOnSameLine_mnemo = Ident::Alias_MultipleStatementsOnSameLine();
my $UselessThen_mnemo = Ident::Alias_UselessThen();
my $GlobalVariables_mnemo = Ident::Alias_GlobalVariables();

my $nb_BadLogicIndents = 0;
my $nb_TabInsideIndentation = 0;
my $nb_BadIndentationStyle = 0;
my $nb_UnexpectedSemicolon = 0;
my $nb_MultipleStatementsOnSameLine = 0;
my $nb_UselessThen = 0;
my $nb_GlobalVariables = 0;



# NOTE : BadLogicIndent known limitations are :
#
# 1 - when expression is on two line with differents indentation => which one should be considered ?
#
# 2 - Some developper are using 4 spaces instead of 2 ...
# 
# 3 - some developpers are using tabs ... 
#
# 4 - should we count one violation per detection ? or one violation for a given bloc of instruction ? 


my $PARENT_REQUIRING_STRICT_INDENTATION = {
	&ModuleKind => 1,
	&ClassKind => 1,
	&MethodKind => 1,
	&BlocKind => 1,
	&BeginKind => 1,
	&IfKind => 1,
	&ThenKind => 1,
	&ElseKind => 1,
	&ElsifKind => 1,
	&WhileKind => 1,
	&ForKind => 1,
	&LoopKind => 1,
	&UntilKind => 1,
	&CaseKind => 1,
	&WhenKind => 1,
	&UnlessKind => 1,
	#&ConditionKind => 1,
	&LambdaKind => 1,
	&BEGINKind => 1,
	&ENDKind => 1,
};

# line_indentation 			= FROM new line TO first non blank character
# expression_indentation 	= FROM new line TO the beginning of the expression containing the concerned statement
# blanked_indentation		= FROM new line TO the concernend statement
# indentation				= FROM end of previous statement TO the concerned statement

# NOTE : the difference between "expression" and "line" is when the beginning of the expression is not on the same line than the considered item ...
# ex          dosomething()
#                     .permform do |a, b|
#               action1()
#               action2()
#             end
#    ==> expression indentation for the item "do" is regarding "dosomething" while line indentation is regarding ".perform()".
my $INDENTATION_REFERENCE = {
	&BlocKind	=> ["expression_indentation", "line_indentation"],
	&BeginKind	=> ["expression_indentation", "line_indentation"],
	&IfKind		=> ["expression_indentation", "line_indentation"],
	&MethodKind	=> ["line_indentation"],
	#~ &ElseKind 	=> "blanked_indentation",
	#~ &ElsifKind 	=> "blanked_indentation",
	#~ &EnsureKind => "blanked_indentation",
	#~ &WhenKind => "blanked_indentation",
	#~ &RescueKind => "blanked_indentation",
	#~ &EnsureKind => "blanked_indentation",
};

# indentation expected with the parent.
# Empty indentation means aligned with parent.
my $INDENTATION_EXPECTED = {
	&ElseKind 	=> "",
	&ElsifKind 	=> "",
	&EnsureKind => "",
	&WhenKind 	=> "",
	&RescueKind => "",
	&EnsureKind => "",
};

# Return the right indentation style, among those specified.
# - the only one specified if any
# - the one who matches the 'end' keyword indentation if any.
# - the first specified if nothing matches
sub get_indentation_reference($) {
	my $node = shift;
	my $kind = GetKind($node);
	my $referencePossible = $INDENTATION_REFERENCE->{$kind};
	my $reference;
	
	if (ref $referencePossible eq "ARRAY") {
		if ((scalar @$referencePossible) == 1) {
			$reference = $referencePossible->[0];
		}
		else {
			my $end_indentation = Ruby::RubyNode::getRubyKindData($node, 'end_indentation');
			if (defined $end_indentation) {
				# try to match the "end" indentation with each possibility
				for my $indent_style (@$referencePossible) {
					my $begin_indent = Ruby::RubyNode::getRubyKindData($node, $indent_style);
					if ((defined $begin_indent) && ($begin_indent eq $end_indentation)) {
						$reference = $indent_style,
						last;
					}
				}
				if (! defined $reference) {
					# none possibility has matched !!!
					# this means that the "end" is badly indented.
					# this cannot help to the decision => consider the first possibility by default.
					$reference = $referencePossible->[0];
					print STDERR "[get_indentation_reference] WARNING: inconsistant 'end' indentation (line ".GetEndline($node).") for '$kind' at line ".GetLine($node)." => using $reference indentation style by default\n";
				}
			}
			else {
				# no "end" indentation to help decision => consider the first possibility by default.
				$reference = $referencePossible->[0];
				print STDERR "[get_indentation_reference] WARNING: missing 'end' indentation for '$kind' at line ".GetLine($node)." => using $reference indentation style by default\n";
			}
		}
	}
	return $reference;
}

sub cb_indentation($$) {
	my $node = shift;
	my $context = shift;
	my $kind = GetKind($node);
	my $line = GetLine($node);
	my $parent = GetParent($node);
	
	# do not check "then node" because it is associated with none statement !!!
	return undef if ($kind eq ThenKind);

	my $indentation = Ruby::RubyNode::getRubyKindData($node, "line_indentation");

	if ($context->[0] eq 'check') {
		# Do not check (and issue a violation) if the node has the same indentation than its previous sibling ...
		#--------------------------------------------------------------------------------------------------------
		if ((!defined $context->[2]->{'parent'}) || ($parent != $context->[2]->{'parent'})) {
			# new context ...
			$context->[2]->{'parent'} = $parent;
			$context->[2]->{'indentation'} = $indentation;
		}
		elsif ($indentation eq $context->[2]->{'indentation'}) {
			# next indentation than previous sibling !
			return undef;
		}
		else {
			# same parent but different indentation.
			$context->[2]->{ "indentation"} = $indentation;
		}
	
		# check "end" keyword indentation if the node is a structure terminated with "end"
		#---------------------------------------------------------------------------------
		my $end_indentation = Ruby::RubyNode::getRubyKindData($node, 'end_indentation');
		if (defined $end_indentation) {
			my $indentation_to_use;
			my $indentation_reference = get_indentation_reference($node);
			if (defined $indentation_reference) {
				$indentation_to_use = Ruby::RubyNode::getRubyKindData($node, $indentation_reference);
			}
			
			if (! defined $indentation_to_use) {
				$indentation_to_use = Ruby::RubyNode::getRubyKindData($node, 'blanked_indentation');
				if (! defined $indentation_to_use) {
					$indentation_to_use = Ruby::RubyNode::getRubyKindData($node, 'line_indentation');
				}
			}
			if ($end_indentation ne $indentation_to_use) {
				Erreurs::VIOLATION($BadLogicIndents_mnemo, "bad indentation of 'end' (line ".(GetEndline($node)||"??").") for '$kind' at line $line ");
			}
		}
	}

	if ((defined $indentation) && (defined $line)) {
		# "Then" is a fictive node => get the real syntaxic parent ... 
		if ($parent->[0] eq ThenKind) {
			$parent = GetParent($parent);
		}
		
		my $parentKind = GetKind($parent);
		my $parentLine = GetLine($parent);
		
		if (exists $PARENT_REQUIRING_STRICT_INDENTATION->{$parentKind}) {
		
			my $parentIndentation;
			my $reference_indentation = get_indentation_reference($parent);
			if (defined $reference_indentation) {
				$parentIndentation = Ruby::RubyNode::getRubyKindData($parent, $reference_indentation);
				if (! defined $parentIndentation) {
					print STDERR "ERROR : missing '$reference_indentation' for kind $parentKind at line $parentLine\n";
				}
			}
			else {
				# by default, work with blanked indenation, and if any, with line indentation
				$parentIndentation = Ruby::RubyNode::getRubyKindData($parent, 'blanked_indentation');
				if (! defined $parentIndentation) {
					$parentIndentation = Ruby::RubyNode::getRubyKindData($parent, 'line_indentation');
				}
			}

			if ((defined $parentIndentation) && (defined $parentLine)) {

				if ($line != $parentLine) {

					# check if child indentation begins with parent one.
					if ($indentation =~ /\A$parentIndentation(.*)/) {
						my $offset = $1;
						my $expected_indentation = $INDENTATION_EXPECTED->{$kind};
						
						# DETECTION mode
						if ($context->[0] eq 'detect') {

							if (! defined $expected_indentation) {
								if (! exists $context->[1]->{$offset} ) {
									$context->[1]->{$offset} = 0;
								}
								$context->[1]->{$offset}++;
							}
						}
						
						# CHECK mode
						else {
							if (! defined $expected_indentation) {
								$expected_indentation = $context->[1];
							}
						
							if ($offset ne $expected_indentation) {
								my $child_indent_length  = () = $indentation =~ /(\s)/g;
								my $parent_indent_length = () = $parentIndentation =~ /(\s)/g;
								Erreurs::VIOLATION($BadLogicIndents_mnemo, "bad indentation ($child_indent_length vs $parent_indent_length for parent) for '$kind' at line $line, whose parent is '$parentKind'");
								$nb_BadLogicIndents++;
							}
						}
					}
					else {
						if ($context->[0] eq 'check') {
							if ((($indentation =~ /\t/) && ($parentIndentation !~ /\t/)) ||
							    (($indentation =~ / /) && ($parentIndentation !~ / /))) {
								Erreurs::VIOLATION($BadLogicIndents_mnemo, "bad (mixed) indentation for '$kind' at line $line, whose parent is '$parentKind'");
							}
							else {
								my $child_indent_length  = () = $indentation =~ /(\s)/g;
								my $parent_indent_length = () = $parentIndentation =~ /(\s)/g;
								Erreurs::VIOLATION($BadLogicIndents_mnemo, "bad (negative) indentation ($child_indent_length vs $parent_indent_length for parent) for '$kind' at line $line, whose parent is '$parentKind'");
							}
							$nb_BadLogicIndents++;
						}
					}
				}
			}
		}
	}
	return undef;
}

sub detectIndentation($) {
	my $root = shift;
	my @context = ('detect', {});
	Lib::Node::Iterate($root, 0, \&cb_indentation, \@context);
	
	my $detected_offset;
	my $nb = 0;
	for my $offset (keys %{$context[1]}) {
		if ($context[1]->{$offset} > $nb) {
			$detected_offset = $offset;
			$nb = $context[1]->{$offset};
		}
	}
	
	if (! defined $detected_offset) {
		$detected_offset = "  "; # default recommanded by style guides.
	}
	
	return $detected_offset;
}

sub CountRuby($$$$)
{
    my ($fichier, $views, $compteurs, $options) = @_;
    
	my $status = 0;
	
	$nb_BadLogicIndents = 0;
	$nb_TabInsideIndentation = 0;
	$nb_BadIndentationStyle = 0;

	my $root = $views->{'structured_code'};
	my $code = \$views->{'code'};

	if ( ! defined $root ) {
		$status |= Couples::counter_add($compteurs, $BadLogicIndents_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $TabInsideIndentation_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	if ( ! defined $code ) {
		$status |= Couples::counter_add($compteurs, $TabInsideIndentation_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	my $offset = detectIndentation($root);
	
	#if the indentation style do not contain tabs, then tab are not expected, so count the number of indentations containing tabs.
	if ($offset !~ /\t/) {
		my $line = 1;
		while ($$code =~ /^ *(\t)?/mg) {
			if (defined $1) {
				$nb_TabInsideIndentation++;
				Erreurs::VIOLATION($TabInsideIndentation_mnemo, "tab inside indentation at line $line");
			}
			$line++;
		}
	}
	
	if ($offset ne "  ") {
		$nb_BadIndentationStyle++;
		Erreurs::VIOLATION($BadIndentationStyle_mnemo, "Bad indentation style, should be two-spaces");
	}
	
	if ($offset ne "  ") {
		print STDERR "Offset indentation is not standard : '$offset'\n";
	}
	
	my @context = ("check", $offset);
	Lib::Node::Iterate($root, 0, \&cb_indentation, \@context);
	
	$status |= Couples::counter_add($compteurs, $BadLogicIndents_mnemo, $nb_BadLogicIndents);
	$status |= Couples::counter_add($compteurs, $TabInsideIndentation_mnemo, $nb_TabInsideIndentation);
	$status |= Couples::counter_add($compteurs, $BadIndentationStyle_mnemo, $nb_BadIndentationStyle);
	
	return $status;
} 

sub CountVariable($$$$) {
    my ($fichier, $views, $compteurs, $options) = @_;
    
	my $status = 0;
	
	$nb_GlobalVariables = 0;

	my $root = $views->{'structured_code'};
	my $code = \$views->{'code'};

	if ( ! defined $code ) {
		$status |= Couples::counter_add($compteurs, $GlobalVariables_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	if ( ! defined $root ) {
		#$status |= Couples::counter_add($compteurs, $TabInsideIndentation_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	# Check global variables
	pos($$code) = undef;
	my %H_global = ();
	my $gvline = 1;
	# search global variables declared in the code, that is whose name is beginning by $, et followed by an assignment ...
	while ($$code =~ /(\n)|(\$\w+)\s*(?:[\+\*\-\/\%\&\|]=|=[^=~])/g) {
		if (defined $1) {
			$gvline++;
		}
		elsif (! exists $H_global{$2}) {
			$H_global{$2} = 1;
			$nb_GlobalVariables++;
			Erreurs::VIOLATION($GlobalVariables_mnemo, "Global variable detected at line $gvline");
		}
	}
	
	$status |= Couples::counter_add($compteurs, $GlobalVariables_mnemo, $nb_GlobalVariables);
	
	return $status;
}

sub CountLinePerLine($$$$)
{
    my ($fichier, $views, $compteurs, $options) = @_;
    
    $nb_UnexpectedSemicolon = 0;
    $nb_MultipleStatementsOnSameLine = 0;
    $nb_UselessThen = 0;
    
    my $status = 0;
    my $code = \$views->{'code'};

	if ( ! defined $code ) {
		$status |= Couples::counter_add($compteurs, $UnexpectedSemicolon_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $MultipleStatementsOnSameLine_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		$status |= Couples::counter_add($compteurs, $UselessThen_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	my $line_number = 1;
	while ($$code =~ /^(.*)$/mg) {
		my $line = $1;
		
		# Count semicolon
		if ($line =~ /;\s*\z/) {
			$nb_UnexpectedSemicolon++;
			Erreurs::VIOLATION($UnexpectedSemicolon_mnemo, "Unexpected semicolon at line $line_number");
		}
		
		# do not check if the line it is a one-line declaration like "def toto; end;"
		#if ($line !~ /\bend\s*;?\s*$/m) {
			
			while ($line =~ /;\s*([^;\s]+)/g) {
				my $item =$1;

				# do not count a violation if the ";" is to separate from end of instruction bloc ...
				next if ($item =~ /\A(end\b|})/);
				$nb_MultipleStatementsOnSameLine++;
				Erreurs::VIOLATION($MultipleStatementsOnSameLine_mnemo, "Several statement on the same line at line $line_number");
			}
		#}
		
		# Count enexpected "then"
		if ($line =~ /(?:\A|[^\.:])\bthen\s*$/m) {
			Erreurs::VIOLATION($UselessThen_mnemo, "Useless 'then' at line $line_number");
			$nb_UselessThen++;
		}
		
		$line_number++;
	}

	$status |= Couples::counter_add($compteurs, $UnexpectedSemicolon_mnemo, $nb_UnexpectedSemicolon);
	$status |= Couples::counter_add($compteurs, $MultipleStatementsOnSameLine_mnemo, $nb_MultipleStatementsOnSameLine);
	$status |= Couples::counter_add($compteurs, $UselessThen_mnemo, $nb_UselessThen);
	
	
	return $status;
}

sub CountBegin() {
	my ($fichier, $views, $compteurs, $options) = @_;

    my $status = 0;
    
	my $KindsLists = $views->{'KindsLists'};

	if ( ! defined $KindsLists ) {
		#$status |= Couples::counter_add($compteurs, $BadLogicIndents_mnemo, Erreurs::COMPTEUR_ERREUR_VALUE );
		
		return $status |= Erreurs::COMPTEUR_STATUS_VUE_ABSENTE;
	}
	
	my $begins = $KindsLists->{&BeginKind};
	
	for my $begin (@$begins) {
		if (GetParent($begin)->[0] ne MethodKind) {
			for my $child (@{GetChildren($begin)}) {
				if (IsKind($child, RescueKind)) {
					print STDERR "Missing contingency method in place of 'begin' at line ".GetLine($begin)."\n";
				}
			}
		}
	}
	
	return $status;
}

1;
