package framework::CMake;

use strict;
use warnings;

use Lib::Sources;
use framework::detections;
use framework::Logs;

use constant TREE_PATH 		=> 0;
use constant TREE_SUBDIRS 	=> 1;
use constant TREE_FILES		=> 2;
use constant TREE_PARENT	=> 3;
use constant TREE_NAME		=> 4;

sub createEmptyNode() {
	return ["<ROOT>", {}, {}, undef, "<ROOT>"];
}

my $H_TREE = createEmptyNode();   # [ path, subdirs, files, parent, name ]

my $CMAKE = {};

my @T_CMakeLists = ();

my @DETECTIONS = ();

#------------------------------ UTILS / LOGS ---------------------------

my $LOG_CATEGORY = {
	"REGISTER" => 1,
	"INCLUDE_RESOLUTION" => 1,
	"DATA_FLOW" => 1,
	"DETECTION" => 1,
	"PARSE" => 1,
};

my $INIT = 0;

sub init() {
	$INIT = 1;
	
	framework::Logs::setLogCategory($LOG_CATEGORY);
}

#-----------------------------------------------------------------------

sub registerFile($) {
	my $fullname = shift;
framework::Logs::Debug("REGISTER $fullname\n", "REGISTER");
	my @dirs = split "/", $fullname;
	my $basename = pop @dirs;
	
	my $path = $H_TREE;
	my $parent = undef;
	my $dir;
	
	# remove "." directories paths ...
	@dirs = grep { !/^\.$/m } @dirs;
	
	for $dir (@dirs) {
		$parent = $path;
		if (exists $path->[TREE_SUBDIRS]->{$dir}) {
			# path already exists ...
			$path = $path->[TREE_SUBDIRS]->{$dir};
		}
		else {
			# create new path ...
			my $subNode = createEmptyNode();
			$subNode->[TREE_PATH] = (defined $path->[TREE_PATH] ? "$path->[TREE_PATH]/$dir" : $dir);
			$path->[TREE_SUBDIRS]->{$dir} = $subNode;
			$path = $path->[TREE_SUBDIRS]->{$dir};
		}
	}
	
	# $path->[TREE_PATH] = $file;
	# $path->[TREE_PATH] =~ s/\/[^\/]*$//;
	my $file = { 'path_node' => $path, 'fullname' => $fullname, 'status' => 'NONE'};
	$path->[TREE_FILES]->{$basename} = $file;
	$path->[TREE_PARENT] = $parent;
	$path->[TREE_NAME] = $dirs[-1] || $H_TREE->[TREE_NAME];  # $dir contains the last directory name of the path ...
	
#~ if (!defined $path->[TREE_NAME]) {
	#~ print "UNDEFINED TRREE NAME !!!\n";
#~ }
	
	if ($basename =~ /\.cmake$/m) {
		if (exists $CMAKE->{$basename}) {
			push @{$CMAKE->{$basename}}, $file;
		}
		else {
			$CMAKE->{$basename} = [$file];
		}
	}
	
	if ($basename eq "CMakeLists.txt" ) {
		push @T_CMakeLists, $file;
	}
	
	return $file;
}

# $from 	: $node of starting point
# $to   	: relative path from starting point node
# -> return : $node of destination path
sub getPath($$) {
	my $from = shift;
	my $to = shift;
	

	
	my @dirs = split "/", $to;
	
	my $pathNode = $from;
	for my $dir (@dirs) {
		next if $dir eq ".";
		
		if ($dir eq "..") {
			$pathNode = $pathNode->[TREE_PARENT]; # parent
			if (!defined $pathNode) {
				framework::Logs::Debug("[framework::CMake::getPath] ERROR : attempt to walk out the tree : parent is null !!\n", "INCLUDE_RESOLUTION");
				return undef;
			}
		}
		else {
			my $path = $pathNode->[TREE_PATH];
			$pathNode = $pathNode->[TREE_SUBDIRS]->{$dir}; # subdir
			if (! defined $pathNode) {
				framework::Logs::Debug("[framework::CMake::getPath] ERROR : attempt to walk out the tree : subdir $dir don't exist in path $path !!\n", "INCLUDE_RESOLUTION");
				return undef;
			}
		}
	}
	return $pathNode;
}

sub applyToCMakeFiles($$;$);
sub applyToCMakeFiles($$;$) {
	my $path = shift;
	my $callback = shift;
	my $level = shift || 0;
	
	# apply callback to files
	for my $filename (keys %{$path->[TREE_FILES]}) {
		if (( $filename eq "CMakeLists.txt") || ($filename =~ /\.cmake$/m)) {
			$callback->($path->[TREE_FILES]->{$filename}, $level+1);
		}
	}
	
	# analyze subdirs of the node
	for my $dir (keys %{$path->[TREE_SUBDIRS]}) {
		applyToCMakeFiles($path->[TREE_SUBDIRS]->{$dir}, $callback, $level+1);
	}
}

#================================ STRIP ================================

my $LINE = 1;

sub parseComment($$) {
	my $content = shift;
	my $file = shift;
	
	my $comment_line = $LINE;
	
	if ($$content =~ /\G\[([^[]*)\[/gc) {
		my $pattern = $1;
		my $lines = "";
		while ($$content =~ /\G(?:(\n)|(\]$pattern\])|(\]|[^\n\]]*))/gc) {
			if (defined $1) {
				$lines .= $1;
				$LINE++;
			}
			elsif (defined $2) {
				return $lines;
			}
		}
		framework::Logs::Error("missing end of comment bracket argument [${pattern}[ opened at line $comment_line for file $file->{'fullname'}.\n");
		return $lines;

	}
	else {
		# swallow until end of line (or end of buffer)...
		$$content =~ /\G[^\n]*(\n|\z)/gc;
		$LINE++;
		return $1;
	}
}

sub parseBracket($$$) {
	my $content = shift;
	my $file = shift;
	my $pattern = shift;
	
	my $comment_line = $LINE;
	
	my $buf = "\[$pattern\[";
	while ($$content =~ /\G(?:(\n)|(\]$pattern\])|(\]|[^\n\]]*))/gc) {
		if (defined $1) {
			$buf .= $1;
			$LINE++;
		}
		elsif (defined $2) {
			$buf .= $2;
			return $buf;
		}
		elsif (defined $3) {
			$buf .= $3;
		}
	}
	framework::Logs::Error("missing end of double bracket opened at line $comment_line for file $file->{'fullname'}.\n");
	return $buf;
}

my %STRINGS = ();
my %STRING_ID = ();
my $CURRENT_ID = 1;

sub parseString($$) {
	my $content = shift;
	my $file = shift;
	
	my $string_line = $LINE;
	my $string_buffer = '"';
	
	while ($$content =~ /\G("|\\"|\\\\|\\|\n|[^"\\\n]+)/gc) {
		if ($1 eq '"') {
			$string_buffer .= $1;
			if (! exists $STRING_ID{$string_buffer}) {
				my $key = "HIGHLIGHT_STRING_$CURRENT_ID";
				$CURRENT_ID++;
				$STRING_ID{$string_buffer} = $key;
				$STRINGS{$key} = $string_buffer;
#print STDERR "STRING = $string_buffer\n";
				return " $key ";
			}
			return $STRING_ID{$string_buffer};
		}
		elsif ($1 eq "\n") {
			$LINE++;
			$string_buffer .= $1;
		}
		else {
			$string_buffer .= $1;
		}
	}
	framework::Logs::Error("unterminated string at line $string_line for file $file->{'fullname'}.\n");
	return " HIGHLIGHT_UNTERMINATED_STRING ";
}
sub strip($$) {
	my $content = shift;
	my $r_file = shift;
	
	my $strip = "";
	
	# all character preceded with \ (except \n) should be considered escaped !
	while ($$content =~ /\G(?:(\n)|(\\[^\n]|\\)|(#)|(")|\[(=*)\[|(\[|[^\n#"\[\\]*))/gc) {
		if (defined $1) {
			$LINE++;
			$strip .= $1;
		}
		elsif (defined $2) {
			# escaped character
			$strip .= $2;
		}
		elsif (defined $3) {
			$strip .= parseComment($content, $r_file);
		}
		elsif (defined $4) {
			$strip .= parseString($content, $r_file);
		}
		elsif (defined $5) {
			$strip .= parseBracket($content, $r_file, $5);
		}
		else {
			$strip .= $6;
		}
	}
#print STDERR "$strip\n";
	return \$strip;
}

sub getStringValue($$) {
	my $strings = shift;
	my $string_id = shift;
	
	my $value = $STRINGS{$string_id};
					
	if (defined $value) {
		$value =~ s/["']//g;
	}
	
	return $value;
}

sub expandString($) {
	my $item = shift;
	
	if (defined $item && $item =~ /\bHIGHLIGHT_STRING_\d+\b/) {
		$item = getStringValue(\%STRINGS, $item);
	}
	return $item;
}

#================================ PARSE ================================

use constant KIND => 0;
use constant PARENT => 1;
use constant CHILDREN => 2;
use constant NAME => 3;
use constant LINE => 4;
use constant DATA => 5;

my $CALLBACK = {
	"function" => \&parseFunction,
	"macro" => \&parseMacro,
	"foreach" => \&parseForeach,
	"add_subdirectory" => \&parseAddSubdirectory,
	"call" => \&parseCall,
};

#my $REG_CALL = qr/(?:\A|\n)[ \t]*(\w+)[ \t]*\(|(\n)/;
my $REG_CALL = qr/^[ \t]*(\w+)[ \t]*\(|(\n)/m;

sub selectCallback($) {
	my $name = shift;

	if (exists $CALLBACK->{$name}) {
		return $CALLBACK->{$name};
	}
	else {
		return $CALLBACK->{'call'};
	}
}

sub printNode($$);
sub printNode($$) {
	my $node = shift;
	my $level = shift;
	
	print "  "x$level.$node->[KIND]." (".($node->[NAME]||"??")." -- ".($node->[LINE]||"??").")\n";
	for my $child (@{$node->[CHILDREN]}) {
		printNode($child, $level+1);
	}
}

sub getParameters($) {
	my $context = shift;
	my $code = $context->{'code'};
	my $r_line = \$context->{'line'};
	my $begin_line = $$r_line;
	
	my @params = ();
	#if ($$code =~ /\G\s*\(\s*/gc) {
		my $level = 0;
		my $arg = "";
		while ($$code =~ /(\()|(\))|(\s)|([^\s\(\)]*)/gc) {
			if (defined $1) {
				$level++;
				$arg .= $1;
			}
			elsif (defined $2) {
				if (!$level) {
					# end of parameters
					if ($arg ne "") {
						push @params, $arg;
					}
					return \@params;
				}
				$level--;
				$arg .= $2;
			}
			elsif (defined $3) {
				if ($3 eq "\n") {
					$$r_line++;
				}
				
				# blank inside enclosed expression do not end the expression
				if ($level > 0) {
					$arg .= $3;
				}
				# blank means end of the expression
				else {
					if ($arg ne "") {
						push @params, $arg;
						$arg = "";
					}
				}
			}
			else {
				$arg .= $4;
			}
		}
	#}
	#else {
	#	print STDERR "CMAKE PARSE ERROR : missing openning parenthese !!!\n";
	#	return [];
	#}
	framework::Logs::Error("missing closing parenthese for parameter list at line $begin_line of file $context->{'file'}->{'fullname'}\n");
}

sub append($$) {
	my $node = shift;
	my $child = shift;
	
	push @{$node->[CHILDREN]}, $child;
	$child->[PARENT] = $node;
}

sub parseFunction($) {
	my $context = shift;
	my $code = $context->{'code'};
	my $r_line = \$context->{'line'};
	
	my $node = [ "function", undef, [], undef, $$r_line, {}];
	
	my $params = getParameters($context);
	$node->[DATA]->{'params'} = $params;
	
	my $name = "";
	if (scalar @$params) {
		$name = shift @$params;
	}
	else {
		framework::Logs::Error("no name for function at line $$r_line of file $context->{'file'}->{'fullname'}\n");
	}
	$node->[NAME] = $name;
	
	while ($$code =~ /$REG_CALL/gc) {
		if (defined $1) {
			#$$r_line++;
			if ($1 eq "endfunction") {
				last;
			}
			append($node, selectCallback($1)->($context, $1));
		}
		else {
			$$r_line++;
		}
	}
	
	return $node;
}

sub parseMacro($) {
	my $context = shift;
	my $code = $context->{'code'};
	my $r_line = \$context->{'line'};
	
	my $node = [ "macro", undef, [], undef, $$r_line, {}];
	
	my $params = getParameters($context);
	$node->[DATA]->{'params'} = $params;
	
	my $name = "";
	if (scalar @$params) {
		$name = shift @$params;
	}
	else {
		framework::Logs::Error("no name for macro at line $$r_line of file $context->{'file'}->{'fullname'}\n");
	}
	$node->[NAME] = $name;
	
	while ($$code =~ /$REG_CALL/gc) {
		if (defined $1) {
			#$$r_line++;
			if ($1 eq "endmacro") {
				last;
			}
			append($node, selectCallback($1)->($context, $1));
		}
		else {
			$$r_line++;
		}
	}
	
	
	return $node;
}

sub parseForeach($) {
	my $context = shift;
	my $code = $context->{'code'};
	my $r_line = \$context->{'line'};
	
	my $node = [ "foreach", undef, [], "", $$r_line, {}];
	
	my $params = getParameters($context);
	$node->[DATA]->{'params'} = $params;
	
	while ($$code =~ /$REG_CALL/gc) {
		if (defined $1) {
			#$$r_line++;
			if ($1 eq "endforeach") {
				last;
			}
			append($node, selectCallback($1)->($context, $1));
		}
		else {
			$$r_line++;
		}
	}
	
	return $node;
}

sub parseAddSubdirectory($) {
	my $context = shift;
	my $code = $context->{'code'};
	my $r_line = \$context->{'line'};
	
	my $node = [ "call", undef, [], "add_subdirectory", $$r_line, {}];
	
	my $params = getParameters($context);
	$node->[DATA]->{'params'} = $params;
	
	return $node;
}

sub parseCall($$) {
	my $context = shift;
	my $name = shift;

	my $code = $context->{'code'};
	my $r_line = \$context->{'line'};
#print STDERR "CALL ($name) at line $$r_line\n";	
	my $node = [ "call", undef, [], $name, $$r_line, {}];
	
	my $params = getParameters($context);
	$node->[DATA]->{'params'} = $params;
	
	# register the call in the context;
	if (! exists $context->{'calls'}->{$name}) {
		$context->{'calls'}->{$name} = [$node];
	}
	else {
		push @{$context->{'calls'}->{$name}}, $node;
	}
	
	return $node;
}

sub parse($) {
	my $context = shift;
	my $code = $context->{'code'};
	my $r_line = \$context->{'line'};

	# 		[ <kind>,	<parent>,	 <children>,	<name>,	<line>,	<DATA> ]
	my $root = ["root",	 undef, 	[],				"root", 1, 		{} ];

	while ($$code =~ /$REG_CALL/gc) {
		if (defined $1) {
			#$$r_line++;
			append($root, selectCallback($1)->($context, $1));
		}
		else {
			$$r_line++;
		}
	}
#printNode($root, 0);
	return $root;
}

#----------------------------------- RESOLVE INCLUDES ----------------------------------
# - $targetRelativePathName is the relative path inside wich we want to find the files. The candidate must match this relative path !!
# - $listOfFiles is the list of file among whose we should choose the most compliant
sub selectFile($$$) {
	my $targetRelativePathName = shift;
	my $basename = shift;
	my $listOfFiles = shift;
framework::Logs::Debug("*** SELECT FILE for $targetRelativePathName/$basename\n", "INCLUDE_RESOLUTION");
	my $candidates = [];
	
	for my $file (@$listOfFiles) {
		# init with path and associated candidate path.
		push @$candidates, [$file->{'path_node'}, $file];
framework::Logs::Debug(" -- candidate : $file->{'path_node'}->[0]/$basename\n", "INCLUDE_RESOLUTION");
	}
	
	$targetRelativePathName =~ s/^\.+\///m;
	
	my @dirs = split "/", $targetRelativePathName;
	
	while (scalar @dirs) {
		
		my $dirname = pop @dirs;
		
		# remove candidate that do not comply !
		for my $candidate (@$candidates) {
			if ($dirname ne $candidate->[0]->[TREE_NAME]) {
				#remove candidate from the list
				$candidate = undef;
			}
		}
		
		my $remainingCandidates = [];
		for my $c (@$candidates) {
			# replace a path with its parent, and the associed candidate file.
			if ((defined $c) && (defined $c->[0]->[TREE_PARENT])) {
				push @$remainingCandidates, [$c->[0]->[TREE_PARENT], $c->[1]];
			}
		}
		
		if (scalar @$remainingCandidates > 0) {
			# continue with remaining
			$candidates = $remainingCandidates;
		}
		else {
			# no remaining. End the loop with previous iteration results (that is the content of $candidates)... 
			last;
		}
		
	}
	
	if (scalar @$candidates == 1) {
		# $node undefined means we have reached the beginning of the target path  !! it's an exact match !!
		return $candidates->[0]->[1];
	}
	elsif (scalar @$candidates > 1) {
		# There are several matches ... return the first one ...
		framework::Logs::Debug("[CMake::selectFile] WARNING : Several matches... returns the first one.\n", "INCLUDE_RESOLUTION");
		return $candidates->[0]->[1];
	}
	else {
		framework::Logs::Debug("[CMake::selectFile] ERROR : none file selected... this case should not occur ... !\n", "INCLUDE_RESOLUTION");
		return undef;
	}
	
}

# STRUCTURE FOR A FILE :{ 
#	'path_node' => ,
#	'fullname' => ,
#  	'status' , 
#  	'context' => { 
#		'code' => , 
#		'line' => ,
#		'calls' => ,
#		'parse' => ,
#		'includes' =>
#		'CallerChainList' =>
# 	}
#
#}

sub getIncludedFile($$) {
	my $include = shift;
	my $fromFile = shift;
	my $fromPath = $fromFile->{'path_node'};
	
	# TODO : should not be restricted to .cmake files ...

	if ($include =~ /\$\{[\s\w]+\}\s$/m) {
		framework::Logs::Debug("[CMake::getIncludedFile] cannot resolve include file : $include\n","INCLUDE_RESOLUTION");
		return undef;
	}

	if ($include !~ /.cmake$/m) {
		$include .= ".cmake";
	}
	
	#my ($basename) = $include =~ /(\w+\.cmake)$/m;
	my $basename = $include;
	$basename =~ s/.*\///m;
	
	my $pathname = $include;
	$pathname =~ s/(^|\/)[^\/]*$//m;
		
	my $candidatesFiles = $CMAKE->{$basename};

	my $includeFile = undef;
		
	if (! defined $candidatesFiles) {
		framework::Logs::Debug("[CMake::getIncludeFile] WARNING : None file $basename in known directory tree ... !!\n", "INCLUDE_RESOLUTION");
	}
	# check path ...
	#elsif ($pathname =~ /^(([^\$]*)\$\{(\w+)\})?(.*)$/m) {
		#~ my $begining;
		#~ my $var;

		#~ # the path contains a variable
		#~ if (defined $1) {
			#~ $begining = $2;
			#~ $var = $3;
			#~ print STDERR "[CMake::getIncludeFile] Included file $include contains a variable !!\n";
			
			#~ # TODO : try to resolve if the variable is known ... 
			#~ # if the known part of the path matches one path associated to candidateFiles, then deduce the value of the variable !!
		#~ }
		
		#~ # the path is understandable ...
		else {
			
			my $destinationPathNode;
			
			# CHECK PRESENCE OF VARIABLE
			if ($pathname =~ /^([^\$]*)\$\{(\w+)\}/m) {
				# the path contains a variable 

#print STDERR "[CMake::getIncludedFile] file $pathname contains a variable !!\n";
		
				# TODO : try to resolve if the variable is known ... 
				# if the known part of the path matches one path associated to candidateFiles, then deduce the value of the variable !!
		
				# keep only the end part without variable ...
				$pathname =~ s/.*\$\{\w+\}\/?//m;
			}
			
			# NO VARIABLE
			#else {
				# compute destination path ...
				$destinationPathNode = getPath($fromPath, $pathname);
			#}
			
			# SELECT FILE
			if (defined $destinationPathNode) {
				# DESTINATION PATH EXISTS
				if (exists $destinationPathNode->[TREE_FILES]->{$basename}) {
					# and contains the required file !
					$includeFile = $destinationPathNode->[TREE_FILES]->{$basename};
					framework::Logs::Debug("[CMake::getIncludeFile] OK : Found $include inside $includeFile->{'path_node'}->[TREE_PATH], and EXPECTED it in $fromPath->[TREE_PATH]/$pathname\n","INCLUDE_RESOLUTION");
					
				}
				else {
					# but do not contain the required file !
					$includeFile = selectFile($pathname, $basename, $candidatesFiles);
					if (defined $includeFile) {
						framework::Logs::Debug("[CMake::getIncludeFile] Found $basename inside $includeFile->{'path_node'}->[TREE_PATH], but EXPECTED it in $fromPath->[TREE_PATH]/$pathname\n", "INCLUDE_RESOLUTION");
					}
					else {
						framework::Logs::Debug("[CMake::getIncludeFile] Can not include $include from $fromPath->[TREE_PATH]\n", "INCLUDE_RESOLUTION");
					}
				}
			}
			else {
				# DESTINATION PATH DON'T EXISTS
				#~ if (scalar @$candidatesFiles == 1) {
					#~ print STDERR "[CMake::getIncludeFile] Nothing corresponding to $include !!!\n";
				#~ }
				#~ else {
					
					# We retain only the end of the path that is pertinent. So remove all what is before the last ../ if any.
					$pathname =~ s/.*(\.\.\/)//g;
					
					$includeFile = selectFile($pathname, $basename, $candidatesFiles);
					if (defined $includeFile) {
						framework::Logs::Debug("[CMake::getIncludeFile] Found $include inside $includeFile->{'path_node'}->[TREE_PATH]\n", "INCLUDE_RESOLUTION");
					}
					else {
						framework::Logs::Debug("[CMake::getIncludeFile] Can not include $include from $fromPath->[TREE_PATH]\n", "INCLUDE_RESOLUTION");
					}
				#~ }
			}
		}
	#}
	
	# path is too complex !
	#~ else {
		#~ print STDERR "[CMake::getIncludeFile] Include path is too complex : $include\n";
	#~ }
	
	if ((defined $includeFile) && ($includeFile == $fromFile) ) {
		framework::Logs::Debug("[getIncludedFile] WARNING : file $fromFile->{'fullname'} can not include itself !\n", "INCLUDE_RESOLUTION");
		$includeFile = undef;
	}
	
	return $includeFile;
}


sub resolveIncludes($$);
sub resolveIncludes($$) {
	my $filenode = shift;
	my $level = shift;

framework::Logs::Debug("RESOLVE INCLUDE FOR $filenode->{'fullname'}\n", "INCLUDE_RESOLUTION");

	for my $nodeCall (@{$filenode->{'context'}->{'calls'}->{'include'}}) {
		my $params = $nodeCall->[DATA]->{'params'};
		my $include = expandString($params->[0]);
		framework::Logs::Debug("    ** INCLUDES $include\n", "INCLUDE_RESOLUTION");
		my $includeFile = getIncludedFile($include, $filenode);
		if (defined $includeFile) {
			framework::Logs::Debug("        ~> FILE INCLUDED by resolution : $includeFile->{'fullname'}\n", "INCLUDE_RESOLUTION");
			push @{$filenode->{'context'}->{'includes'}}, $includeFile
		}
	}
}

#----------------------------------- ANALYZE ----------------------------------

sub printStack($;$) {
	my $stack = shift;
	my $last = shift;
	
	for my $trace (@$stack) {
		framework::Logs::Debug("... from $trace\n", "DETECTION");
	}
	framework::Logs::Debug("... from $last\n", "DETECTION") if $last;
}

sub initFileAnalyse() {
	# %STRINGS = ();
	# %STRING_ID = ();
	# $CURRENT_ID = 1;
	
	$LINE = 1;
}


# Search upstream initialisation of a variable.
# - $node is a node of the tree representation on the code inside the file
# - $var is the name of the variable we want to know the upstream initialisation before the node.
sub findUpstreamVariable($$);
sub findUpstreamVariable($$) {
	my $node = shift;
	my $var = shift;
	
	my $parent = $node->[PARENT];
	
	my $ret = { 'functions' => [], 'deps' => [], 'vars' => [] };
	
	my $found = 0;
	while ($parent->[KIND] ne "root") {
#print STDERR "    UPSTREAM for $var : $parent->[KIND] at line $parent->[LINE]\n";

		# the requested variable takes its values inside a loop ...
		# FIXME : as for functions, the first parameter of the foreach should be considered as the name of the foreach (all other parameters are data ...)
		if ($parent->[KIND] eq "foreach") {
			
			my $params = $parent->[DATA]->{'params'};
			if ($params->[0] eq $var) {
				my $idx = scalar @$params - 1;
				while ($idx) {
					my $item = expandString($params->[$idx]);
					if ($item =~ /^\w+$/) {
						push @{$ret->{'deps'}}, [$item, $parent->[LINE], "foreach at line $parent->[LINE]"];
					}
					elsif ($item =~ /^\$\{(\w+)\}$/) {
						# FIXME : should search upstream ...
						framework::Logs::Debug("  Requested variable '$var' take this variable value '$item' in a foreach at line $parent->[LINE]\n", "DATA_FLOW");
						my $var_ret = findUpstreamVariable($parent, $1);
						if (scalar @{$var_ret->{'functions'}}) {
							my $func = $var_ret->{'functions'}->[0];
							push @{$ret->{'functions'}}, [$func->[0], $func->[1], $func->[2], "Parameter index $func->[2] of function $func->[0] declared at line $func->[1]" ];
						}
					}
					else {
						framework::Logs::Debug("  Requested variable '$var' take this complex value '$item' in a foreach at line $parent->[LINE]\n", "DATA_FLOW");
					}
					$idx --;
				}
				
				# the initialisation of the requested variable has been resolved ... 
				$found = 1;;
			}
		}
		# the requested variable takes its value from a function parameter ...
		elsif (($parent->[KIND] eq "function") || ($parent->[KIND] eq "macro")) {
			my $params = $parent->[DATA]->{'params'};
			my $idxParam = 0;
			my $found = 0;
			for my $param (@$params) {
				my $item = expandString($param);
				if ($item eq $var) {
					push @{$ret->{'functions'}}, [$parent->[NAME], $parent->[LINE], $idxParam, "Parameter index $idxParam of function $parent->[NAME] declared at line $parent->[LINE]" ];
					
					# the initialisation of the requested variable has been resolved ... 
					$found = 1;
					last;
				}
				$idxParam++;
			}
		}

		if ($found) {
			last;
		}

		# next parent iteration
		$parent = $parent->[PARENT];
	}
	return $ret;
}


sub searchDataFlowInFile($$$$$$);
sub checkCallerChainList($$$$);
	
sub checkCallerChainList($$$$) {
	my $CallerChainList = shift;
	my $key = shift;
	my $file =shift;
	my $traceStack = shift;
	
	# file inherit from its include file. So update its own CallerChainList with those of its include
	if (! exists $file->{'context'}->{'CallerChainList'}->{$key}) {
		$file->{'context'}->{'CallerChainList'}->{$key} = [];
	}
	push @{$file->{'context'}->{'CallerChainList'}->{$key}}, @{$CallerChainList->{$key}};
	
	for my $callerKey (@{$CallerChainList->{$key}}) {
		my ($function, $argIdx) = split /#/, $callerKey;
		searchDataFlowInFile($function, $argIdx, undef, $file, 0, $traceStack);
		
		checkCallerChainList($CallerChainList, $callerKey, $file, $traceStack);
	}
}

# - $functionName : a function that takes a dependency in parameter
# - $argIdx : the index of the dependency parameter 
# - $file   : the file inside wich we are searching
# - $traceStack : a stack for the logs ...
sub searchDataFlowInFile($$$$$$) {
	my $functionName = shift;
	my $argNameIdx = shift;
	my $argVersionIdx = shift;
	my $file = shift;
	my $searchInsideIncludes = shift;
	my $traceStack = shift;
	
	my $key = "$functionName#$argNameIdx";
	
	# If a callerChainList exists for the key, means the file has already been treated !!
	if (! exists $file->{'context'}->{'CallerChainList'}->{$key}) {
		
		$file->{'context'}->{'CallerChainList'}->{$key} = [];

framework::Logs::Debug("search data flow in FILE $file->{'fullname'}, FUNCTION $functionName, ARG $argNameIdx, with searchInsideIncludes = $searchInsideIncludes\n", "DATA_FLOW");
		# search all calls of the function $functionName
		for my $nodeCall (@{$file->{'context'}->{'calls'}->{$functionName}}) {

			# get the dependency (the asked parameter)
			my $dep = $nodeCall->[DATA]->{'params'}->[$argNameIdx];
		
			if (defined $dep) {
				$dep =~ s/\s+//g;
			
				# Case it is a strings ==> get the value
				if ($dep =~ /\bHIGHLIGHT_STRING_\d+\b/) {
					$dep = getStringValue(\%STRINGS, $dep);
				}

				# GET VERSION 
				my $version;
				if (defined $argVersionIdx) {

					$version = expandString($nodeCall->[DATA]->{'params'}->[$argVersionIdx]);

					if (defined $version) {
						if ($version =~ /\$\{(\w+)\}/m) {
							framework::Logs::Debug("Version $version can not be resolved from variable\n", "DATA_FLOW");
							$version = undef;
						}
						elsif ($version !~ /\d|\./) {
							$version = undef;
						}
					}
				}

				push @$traceStack, "Arg $dep (version ".($version||"??").") of function $functionName called at line $nodeCall->[LINE] of file $file->{'fullname'}";
				
				# CHECK NAME 
				if ($dep =~ /^\$\{(\w+)\}$/m) {
					# The dependency is inside a variable
					my $ret = findUpstreamVariable($nodeCall, $1);
				
					# Resolution of the variable has retrived following dependencies
					for my $dep (@{$ret->{'deps'}}) {
						framework::Logs::Debug("--> FOUND DEPENDENCY $dep->[0] (version ".($version||"??").") at line $dep->[1]\n", "DETECTION");
						#my $itemDetection = framework::detections::getEnvItem($dep->[0], $dep->[2], undef, 'CMake', $file->{'fullname'}, undef);
						my $itemDetection = framework::detections::createItemDetectionWithoutDescription($dep->[0], undef, $version, $framework::dataType::STATUS_DISCOVERED, $file->{'fullname'}, "CMake");
						if (defined $itemDetection) {			
							push @DETECTIONS, $itemDetection;
							
						}
						printStack($traceStack,$dep->[2]);
					}
				
					# Resolution of the variable ask to investigate some function calls :
					for my $func (@{$ret->{'functions'}}) {
					
						# Associate the functions. We should be able to know what functions are calling the function given in parameter, that we know she declares a dependency ...
						# record $func->[0] as a function that is calling $functionName and then provide a dependency from its own parameters ...
						push @{$file->{'context'}->{'CallerChainList'}->{$key}}, "$func->[0]#$func->[2]";
					
						# search calls of $func to detect dependency at argument $func->[2]
						framework::Logs::Debug("--> SHOULD SEARCH Calls of $func->[0], parameter $func->[2]\n", "DATA_FLOW");
						searchDataFlowInFile($func->[0], $func->[2], undef, $file, 0, $traceStack);
					
						# record these functions for calling context ...
						#push @{$file->{'context'}->{'FunctionsWithDeps'}}, $func;
					}
				}
				elsif ($dep =~ /^(\w+)$/m) {
					# The dependency readable
					framework::Logs::Debug("--> FOUND DEPENDENCY : $1 (version ".($version||"??").") !!\n", "DETECTION");
					my $itemDetection = framework::detections::createItemDetectionWithoutDescription($dep, undef, $version, $framework::dataType::STATUS_DISCOVERED, $file->{'fullname'}, "CMake");
					if (defined $itemDetection) {	

						push @DETECTIONS, $itemDetection;
					}
					printStack($traceStack);
				}
				else {
					framework::Logs::Debug("[CMAKE::findPackageDataFlow] WARNING : unable to expand parameter $dep in $nodeCall->[NAME] at line $nodeCall->[LINE]\n", "DATA_FLOW");
				}
					
				pop @$traceStack;
			}
			else {
				framework::Logs::Debug("[CMAKE::findPackageDataFlow] ERROR : missing parameter for $nodeCall->[NAME] at line $nodeCall->[LINE]\n", "DATA_FLOW");
			}
		}
	}
	else {
framework::Logs::Debug("(!)search data flow already DONE for FILE $file->{'fullname'}, FUNCTION $functionName, ARG $argNameIdx\n", "DATA_FLOW");
	}

	if ($searchInsideIncludes) {
		# search inside include files
		for my $include (@{$file->{'context'}->{'includes'}}) {
			# search calls of $functionName inside include file.
			searchDataFlowInFile($functionName, $argNameIdx, $argVersionIdx, $include, 1, $traceStack);

			# if the include file defines functions that calls $functionName and the dep is an argument of these caller function,
			# then we should check in the file if it uses these caller function, and if any, the callers of the caller ...
			checkCallerChainList($include->{'context'}->{'CallerChainList'}, $key, $file, $traceStack);
		}
	}
}

sub searchDataFlowInFileHierarchy($$$$$);
sub searchDataFlowInFileHierarchy($$$$$) {
	my $functionName = shift;
	my $argNameIdx = shift;
	my $argVersionIdx = shift;
	my $file = shift;
	my $traceStack = shift;

	searchDataFlowInFile($functionName, $argNameIdx, $argVersionIdx, $file, 1, $traceStack);

	#~ # search inside include files
	#~ for my $include (@{$file->{'context'}->{'includes'}}) {
		#~ searchDataFlowInFile($functionName, $argIdx, $include, $traceStack);
		
		#~ # retrieves functions providing dependencies from include file.
		#~ my $provider_functions = $include->{'context'}->{'CallerChainList'};
	#~ }
	
	
}

# TODO :
# - résoudre les appel à add_subdirectory
# - résoudre les inclusions de fichier CMake
# - lorsque un package est dans une variable, voir si on peut résoudre la variable ... ?? (mmmh ... probablementy trop dur).
# liens de travail:
# https://github.com/daminetreg/dbus/blob/master/cmake/dbus/CMakeLists.txt
# https://www.incredibuild.com/blog/working-with-cmake-dependencies
# https://github.com/Kitware/CMake/tree/master
# https://cmake.org/cmake/help/v3.0/manual/cmake-commands.7.html
sub parseFiles($$;$);
sub parseFiles($$;$) {
	my $file = shift;
	my $level = shift;
	my $content = shift;
	
	my $indentation = ("    " x $level);
	my $pathNode = $file->{'path_node'};
	
#print STDERR "${indentation}ANALYZE $pathNode->[TREE_PATH]\n";
framework::Logs::Debug("${indentation}ANALYZE $file->{'fullname'}\n", "PARSE");

	if ($file->{'status'} eq 'PENDING') {
		framework::Logs::Error("ERROR : ERROR : circular reference detected !!\n");
		return;
	}

	if ($file->{'status'} eq 'NONE') {

		$file->{'status'} = 'PENDING';

		if (! defined $content) {
			$content = Lib::Sources::getFileContent($file->{'fullname'});
		}
		
		if (! defined $content) {
			framework::Logs::Warning("Unable to read $file->{'fullname'} for CMake dependencies inspection purpose.\n");
			return undef;
		}
		
		# init for each file...
		initFileAnalyse();
		
		# strip
	    my $strippedContent = strip($content, $file);
		
		# parse
		my $context = { 'file' => $file, 'code' => $strippedContent, 'line' => 1, 'calls' => {}, 'includes' => [] };
		$context->{'parse'}= parse($context);
		$file->{'context'} = $context;
		
		return;
	}
	
	$file->{'status'} = "ANALYZED";
}



# analyze a NODE.
#~ sub analyzePath($);
#~ sub analyzePath($) {
	#~ my $path = shift;
	
	#~ # analyze the CMakeLists.txt file present in the node
	#~ if (exists $path->[TREE_FILES]->{'CMakeLists.txt'}) {
		#~ my $file = $path->[TREE_FILES]->{'CMakeLists.txt'};
		#~ if ($file->{'status'} ne "ANALYZED") {
			#~ analyseCMakeFile($file, 0);
		#~ }
	#~ }
	
	#~ # analyze subdirs of the node
	#~ for my $dir (keys %{$path->[TREE_SUBDIRS]}) {
		#~ analyzePath($path->[TREE_SUBDIRS]->{$dir});
	#~ }
#~ }

sub analyzeCMake($) {
	my $params = shift;
	
	init();
	
	# register all files to create the tree...
	for my $dir (keys %{$params->{'directories'}}) {
		for my $file (keys %{$params->{'directories'}->{$dir}}) {
			registerFile("$dir/$file");
		}
	}
	
	#~ # analyze *.cmake
	#~ for my $name (keys %$CMAKE) {
		#~ for my $file (@{$CMAKE->{$name}}) {
			#~ if ($file->{'status'} ne "ANALYZED") {
				#~ analyseCMakeFile($file, 0);
			#~ }
		#~ }
	#~ }
	
	applyToCMakeFiles($H_TREE, \&parseFiles);
	
	applyToCMakeFiles($H_TREE, \&resolveIncludes);
	
	my $traceStack = [];
	for my $CMakeLists (@T_CMakeLists) {
		searchDataFlowInFileHierarchy("find_package", 0, 1, $CMakeLists, $traceStack);
	}
	
	# begin analyze of all CMakeLists.txt by the root node ...
	#analyzePath($H_TREE);

	return \@DETECTIONS;
}

####################### TEST ACCESSORS #################################

sub test_init() {
	@DETECTIONS = ();
	$H_TREE = createEmptyNode();   # [ path, subdirs, files, parent, name ]
	$CMAKE = {};
	@T_CMakeLists = ();
}

sub get_DETECTIONS() {
	return \@DETECTIONS;
}

1;
