# Swift Package Manager
package framework::SPM;

use strict;
use warnings;

use framework::dataType;
use framework::detections;
use framework::Logs;
use Lib::Sources;
use File::Basename;
use File::Spec;
use JSON;
use Lib::LongPathUtils;

BEGIN {
	if ($^O eq 'MSWin32') {
		require Win32::LongPath;
		Win32::LongPath->import;
	}
}

my $DEBUG = 0;

my %views = ();
my @dependencies = ();

sub init() {
	%views = ();
	@dependencies = ();
}
#-----------------------------------------------------------------------
#                       DEPENDENCIES ROUTINES 
#-----------------------------------------------------------------------

sub addDependency($) {
	my $dep = shift;
	if (ref $dep eq "HASH") {
		for my $key (keys %$dep) {
			if ($dep->{$key}) {
				$dep->{$key} =~ s/["']//g;
			}
		}
		push @dependencies, $dep;
		framework::Logs::Debug("  --> FOUND DEPENDENCY : " . ($dep->{'group'} || "") . ":" . ($dep->{'name'} || "") . ":" . ($dep->{'version'} || "") . "\n");
	}
	elsif (ref \$dep eq "SCALAR") {
		$dep =~ s/["']//g;
		framework::Logs::Debug("  --> FOUND DEPENDENCY : $dep\n");
		my @data = split ":", $dep;
		push @dependencies, { 'group' => $data[0], 'name' => $data[1], 'version' => $data[2] };
	}
	else {
		framework::Logs::Warning("Unknown version format\n");
	}
}

# dependencies: [
	# Dependencies declare other packages that this package depends on.
	# .package(url: "https://url/of/another/package.git", from: "1.0.0"),
	# .package(url: "https://github.com/johnsundell/files.git", .exact("4.0.0")
	# .package(name: "AppNexusSDK", url: "https://github.com/appnexus/mobile-sdk-ios", .exact("8.2.0"))
	# .package(name: "GoogleUserMessagingPlatform",url: "https://github.com/googleads/swift-package-manager-google-user-messaging-platform.git", "1.1.0"..<"3.0.0")
	# .package(path: "path/to/a/local/package/", "1.0.0"..<"2.0.0")]
	# .package(path: "../CalendarKit")
sub parseBlocDependencies($$) {
	my $buf_dep = shift;
	my $dependencies = shift;

	while ($buf_dep =~ /^\s*\.package\s*\((.*?)\)/simg) {
		my $match = $1;
		my $name;
		my $location;
		my $version;
		# component name
		if ($match =~ /\bname\s*\:\s*["'](.*?)["']/) {
			$name = $1;
		}
		# component location
		if ($match =~ /(?:\bpath|\burl)\s*\:\s*["'](.*?)["']/) {
			$location = $1;
		}
		if ($location =~ /.*[\\\/]([\w\-\_]+)(?:\.git)?/) {
			$name = $1;
		}
		# component version
		if ($match =~ /(\bfrom|\.exact|\brevision|\bbranch)\s*\:?\s*\(?["'](.*?)["']\)?/) {
			my $typeVar = $1;
			if ($typeVar eq 'from') {
				$version = '^' . $2;
			}
			elsif ($typeVar eq '.exact') {
				$version = $2;
			}
			elsif ($typeVar eq 'revision' || $typeVar eq 'branch') {
				$version = '';
			}
		}
		# range of version without prefix...
		# .package(name: "name",url: "url", "1.1.0"..<"3.0.0")
		elsif ($match =~ /.*?\,.*?\,(.*)/s) {
			$version = $1;
		}

		$dependencies->{$name}->{'group'} = $location if defined $location;
		$dependencies->{$name}->{'version'} = $version if defined $version;
	}
	return $dependencies;
}

sub parseBlocTargets($$) {
	my $buf_target = shift;
	my $dependencies = shift;
	my %correspondingProductName;
	foreach my $dep (@{$dependencies}) {
		if ($buf_target =~ /^\s*\.product\s*\(\s*name\s*\:\s*\"([^"]+)\"\s*\,\s*package\s*\:\s*\"$dep->{'name'}\"/im) {
			$correspondingProductName{$dep->{'name'}} = $1;
		}
	}
	return \%correspondingProductName;
}

sub parsePackageSwift($) {
	my $buf = shift;
	my $dependencies;

	if ($$buf =~ /^\s*dependencies\s*\:\s*\[([^\]]+)/m) {
		$dependencies = parseBlocDependencies($1, $dependencies);
	}

	# persist dependencies
	foreach my $dep (keys %{$dependencies}) {
		addDependency({ 'group' => $dependencies->{$dep}->{'group'}, 'name' => $dep, 'version' => $dependencies->{$dep}->{'version'} });
	}
}

sub parsePackageResolved($) {
	my $buf = shift;
	my %H_dependencies_resolved;

	my $json = JSON->new->allow_nonref;
	my $decoded_json_resolved = $json->decode($$buf);

	# VERSION 1
	if (exists $decoded_json_resolved->{'object'}->{'pins'}) {
		foreach my $component (@{$decoded_json_resolved->{'object'}->{'pins'}}) {
			if (exists $component->{'package'}) {
				$H_dependencies_resolved{$component->{'package'}}{'group'} = $component->{'repositoryURL'};
				$H_dependencies_resolved{$component->{'package'}}{'version'} = $component->{'state'}->{'version'};
			}
		}
	}
	# VERSION 2
	elsif (exists $decoded_json_resolved->{'pins'}) {
		foreach my $component (@{$decoded_json_resolved->{'pins'}}) {
			if (exists $component->{'identity'}) {
				$H_dependencies_resolved{$component->{'identity'}}{'group'} = $component->{'location'};
				$H_dependencies_resolved{$component->{'identity'}}{'version'} = $component->{'state'}->{'version'};
			}
		}
	}
	else {
		framework::Logs::Warning("[SwiftPM] version of Package.resolved not supported\n");
	}

	# persist dependencies
	foreach my $dep (keys %H_dependencies_resolved) {
		addDependency({ 'group' => $H_dependencies_resolved{$dep}{'group'}, 'name' => $dep, 'version' => $H_dependencies_resolved{$dep}{'version'} });
	}
}

# Swift Package Manager
sub package_swift($$$) {
	my $files_packages = shift;
	my $package_DB = shift;
	my $H_DatabaseName = shift;
	my @itemDetections = ();
	my $correspondingProductName;

	foreach my $file_package (@{$files_packages}) {
		init();
		framework::Logs::printOut("SWIFT PACKAGE : $file_package\n");
		my $buf = Lib::Sources::getFileContent($file_package);
		if (!defined $buf) {
			framework::Logs::Warning("Unable to read $file_package for framework inspection purpose.\n");
		}
		else {
			%views = ('text' => $$buf);
			my $filePackageSwift;
			if (basename($file_package) eq 'Package.swift') {
				parsePackageSwift(\$views{'text'});
				$filePackageSwift = $file_package;
			}
			elsif (basename($file_package) eq 'Package.resolved') {
				parsePackageResolved(\$views{'text'});
				$filePackageSwift = dirname($file_package) . '/Package.swift';
			}

			# if Package.swift exists in same directory
			if (defined $filePackageSwift) {
				if (Lib::LongPathUtils::PathExists($filePackageSwift)) {
					if (basename($file_package) eq 'Package.resolved') {
						$buf = Lib::Sources::getFileContent($filePackageSwift);
						if (!defined $buf) {
							framework::Logs::Warning("[framework::SPM] Unable to read $filePackageSwift for framework inspection purpose.\n");
							return undef;
						}
					}
				}
				else {
					framework::Logs::Warning("[framework::SPM] Search of component name not done: $filePackageSwift not exists\n");
				}
			}

			# change name of component in @dependencies
			if ($$buf =~ /^\s*targets\s*\:\s*\[(.*)\]/sm) {
				my $buf_targets = $1;
				$correspondingProductName = parseBlocTargets($buf_targets, \@dependencies);
			}

			for my $dep (@dependencies) {
				my $spmItem;
				if (exists $correspondingProductName->{$dep->{'name'}}) {
					$spmItem = framework::detections::getEnvItem($correspondingProductName->{$dep->{'name'}}, $dep->{'version'}, $package_DB, 'SPM', $file_package, $H_DatabaseName);
				}
				else {
					$spmItem = framework::detections::getEnvItem($dep->{'name'}, $dep->{'version'}, $package_DB, 'SPM', $file_package, $H_DatabaseName);
				}

				if (defined $spmItem) {
					# if groupId has been captured, then re-build the framework name as this : <groupId>:<artifactId>
					if (defined $dep->{'group'}) {
						# in Swift, group is an url from github
						$dep->{'group'} =~ s/^https?\:\/\///m;
						$spmItem->{'framework_name'} = $dep->{'group'} . ":" . $spmItem->{'framework_name'};
					}

					push @itemDetections, $spmItem;
				}
			}
			framework::Logs::Debug("-- End SWIFT PACKAGE detection-- \n");
		}
	}

	return \@itemDetections;
}

1;
