package Lib::Sources;

use strict;
use warnings;
use AnalyseOptions;
use Encode qw (encode);
use Lib::LongPathUtils;
use Cwd;

BEGIN {
	if ($^O eq 'MSWin32') {
		require Win32::LongPath;
		Win32::LongPath->import;
	}
}

my $ROOT_DIR = undef;

sub setRootDir($) {
	$ROOT_DIR = shift;
}

sub initRootDir() {
	$ROOT_DIR = setRootDir(getcwd);
}

sub getFileContent($) {
	my $filename = shift;

	my $ret;
	my $OUTPUT_FILE_HANDLER;
	if ($^O eq 'MSWin32') {
		$ret = Win32::LongPath::openL(\$OUTPUT_FILE_HANDLER, '<', $filename);
	}
	else {
		$ret = open $OUTPUT_FILE_HANDLER, "<$filename";
	}

	if (! defined $ret ) {
		return undef;
	}

	local $/=undef;
	my $buf = <$OUTPUT_FILE_HANDLER>;
	close $OUTPUT_FILE_HANDLER;

	return \$buf;
}

sub getSourceRootDir($) {
	my $fileList = shift;

	my %H_reps = ();

	for my $file (@$fileList) {
		my @subreps = split /[\/\\]/, $file;
		# remove the file name.
		pop @subreps;

		# extract the base of the path: ., .., <ident>, <drive>: (windows), <empty string> (root linux)
		my $base;
		if (scalar @subreps > 0) {

			if ((scalar @subreps > 1) && ($subreps[0] eq '') && ($subreps[1] eq '')) {
				# case of windows network path  \\<network>\ ...
				#    ==> @subreps begins with two empty strings.
				shift @subreps;
				shift @subreps;
				# add a single /. The second one will be added as separator when concatenating with the rest of the path !
				$base ="/";
			}
			else {
				$base = shift @subreps;
			}

		}
		else {
			$base = '.';
		}

		# init the path related to the base ...
		my $tree = $H_reps{$base};
		if (! defined $tree) {
			# THE FILE IS IN AN UNKNOW PATH
			# Associate a hash with each sub rep (for late exploitation, hash existing test is faster than string comparison) !!
			my @hreps = ();
			for my $rep (@subreps) {
				push @hreps, { $rep => 1};
			}
			$H_reps{$base} = \@hreps;
			next;
		}

		# THE FILE IS IN AN KNOWNED PATH
		# compare the path of the file to the base tree ... subrep by subrep
		# and truncate the base tree where it differs from the file path !
		my $idx = 0;
		my $nb_subreps = scalar @subreps;
		for my $hrep (@$tree) {
			# check if the idx'th subrep of the file is the same of the idx'th subrep of the tree:
			if ( ($idx == $nb_subreps) || (! exists $hrep->{$subreps[$idx]})) {
				# the path element of the file differs from the base element.
				# So the base element do not belong to a common base of all files directories.
				# So all base element from the index $idx are removed !!
				splice @$tree, $idx;
			}
			$idx++;
		}
	}

	# build sources directories (generally one, but it can be several if the list of files is build manually by user !!
	my @tabSrcDir = ();
	for my $base (keys %H_reps) {
		my $tree = $H_reps{$base};
		#my $sourceDir = $base."/";
		my $sourceDir = $base;
		for my $hrep (@$tree) {
			my @tab = keys %$hrep;
			$sourceDir .= "/".(shift @tab);
		}
		push @tabSrcDir, $sourceDir;
		print "Source directory detected : $sourceDir\n";
	}
	return \@tabSrcDir;
}


# get a list of base directories corresponding to a list of files.
# There can be several directories if there is no common root directory.
sub getSourceDir($$) {
	my $fileList = shift;
	my $options = shift;

	# return undef, '' or a directory name terminated with '/' corresponding to the directory specified with "--dir-source" ...
	my $SourceDir = AnalyseOptions::GetSourceDirectory($options);
	my $tabSourceDir ;
	if ((!defined $SourceDir) || ($SourceDir eq '')) {
		$tabSourceDir = getSourceRootDir($fileList);
	}
	else {
		print "Source directory detected : $SourceDir\n";
		$tabSourceDir = [$SourceDir];
	}

	return $tabSourceDir;
}

sub canonizePath($) {
	my $path = shift;

	$$path =~ s/^\\\\/:RESO:/g;
	
	# replace all "\" by "/"
	$$path =~ s/\\/\//g;

	# replace multiple "/" by single "/"
	if ( $$path =~ /^\/\//m) {
		# NETWORK PATH
		$$path =~ s/\/+/\//g;
		# the beginning // has been replaced with /.
		# So add the missing / to build the // necessary for network path syntax.
		$$path = "/".$$path;
	}
	else {
		# OTHER PATH
		$$path =~ s/\/+/\//g;
	}
	# remove trailing "/"
	$$path =~ s/\/$//;
	
	$$path =~ s/^:RESO:/\\\\/g;
}

sub simplifyCanonizedPath($) {
	my $path = shift;
	
	# remove /.
	# rep1/./rep2/./rep3   =>  rep1/rep2/rep3
	my $tocheck = 1;
	while ($tocheck) {
		# remove trailing /.
		$path =~ s/\/\.$//m;
		
		# remove /. between two directorlies
		$tocheck = $path =~ s/\/\.\//\//m;
	}
	
	# remove ../
	my $tocheck1 = 1;
	my $tocheck2 = 1;
	while ($tocheck1 || $tocheck2) {
		#  ../rep1             ==>  rep1
	    #  ./rep1              ==>  rep1
	    $path =~ s/^(\.\.?\/)+//m;
	    
		#  rep1/rep2/../rep3   ==>  rep1/rep3	    
	    #  c:/rep1/../rep2     ==>  c:/rep2
	    #  /rep1/../rep2       ==>  /rep2
		$tocheck1 = $path =~ s/\/[^\/]+\/\.\.\//\//m;

		#  rep1/../rep2        ==>  rep2
		$tocheck2 = $path =~ s/^[^\/]+\/\..\///m;

	}
	
	#  ../rep1             ==>  rep1
	#  ./rep1              ==>  rep1
	$path =~ s/^(\.\.?\/)+//m;
	
	$path =~ s/^(\w:\/)/uc($1)/ge;
	
	return $path;
}


sub canonizeListOfPaths($) {
	my $list = shift;
	my @canons = (split ",", $list);
	for my $path (@canons) {
		canonizePath(\$path);
	}
	return join ",", @canons;
}

sub isAbsolutePath($) {
	my $path = shift;
	if ($path =~ /^(?:\\\\|([a-zA-Z]:)?[\\\/])/) {
		return 1;
	}
	return 0;
}

sub buildAbsoluteCanonizedPath($) {
	my $dir = shift;
	
	my $absolute = $dir;
	if (! isAbsolutePath($dir)) {
		$absolute = $ROOT_DIR."/$dir";
	}
	canonizePath(\$absolute);
	return simplifyCanonizedPath($absolute);
}

#-----------------------------------------------------------------------



my $findFiles_fileList;
my $findFiles_filePattern;

my $R_IGNORED = undef;
my $R_IGNORED_PATH = undef;
my $R_IGNORED_ABSOLUTE_PATH = undef;
my $R_IGNORED_RELATIVE_VS_ABSOLUTE_PATH = undef;

sub buildIgnoreRE($) {
	my $ignore = shift;
	my $ignoreCanon = canonizeListOfPaths($ignore);
	$ignoreCanon =~ s/,/|/g;
	$ignoreCanon &&= '^('.$ignoreCanon.')$';
	if ($ignoreCanon) {
		eval {
			$R_IGNORED = qr($ignoreCanon);
		};
		if ($@) {
			print STDERR "[Lib::Sources] ERROR : Bad regexp for file exclusion pattern : $ignore\n";
		}
	}
}

sub buildIgnorePathsRE($) {
	my $ignorePaths = shift;
	$ignorePaths =~ s/,/|/g;
	#my $ignoreCanonPaths = canonizeListOfPaths($ignorePaths);
	#$ignoreCanonPaths =~ s/,/|/g;
	
	my @paths = split /\|/, $ignorePaths;
	for my $path (@paths) {
		$path =~ s/(^\w:)/uc($1)/e;
		$path =~ s/\\\\/\//g;
		$path =~ s/\/+/\//g;
    
	}
	$ignorePaths = join('|', @paths);
	
	# paths can be absolute or relative to the entry point of the analyze (depending on how the analyze has been launched).
	# so, allow left alignment on beginning or on path separator.
	$ignorePaths &&= '^('.$ignorePaths.')$';

	if ($ignorePaths) {
		eval {
			$R_IGNORED_PATH = qr($ignorePaths);
		};
		if ($@) {
			print STDERR "[Lib::Sources] ERROR : Bad regexp for path exclusion pattern : $ignorePaths\n";
			$R_IGNORED_PATH = undef;
		}
	}
}

sub init($) {
	my $options = shift;

	initRootDir();
	
	if (defined $options->{'--ignore'}) {
		buildIgnoreRE($options->{'--ignore'});
	}

	if (defined $options->{'--ignorePaths'}) {
		buildIgnorePathsRE($options->{'--ignorePaths'});
	}
}


sub isIgnored($) {
	my $name = shift;
	return $R_IGNORED && $name =~ $R_IGNORED;
}

sub isIgnoredPath($) {
	my $name = shift;

	#~ # $name is a relative path
    if ($R_IGNORED_PATH && $name =~ $R_IGNORED_PATH) {
		return 1;
}
	return 0;
}

sub findFiles_triggerAdd($) {
	my $file = shift;
	push @$findFiles_fileList, $file;
}

sub findFiles_fileMatch($) {
	my $file = shift;

	if (!defined $findFiles_filePattern) {
		return 1;
	}

	if ($file =~ /^$findFiles_filePattern$/m) {
		return 1;
	}
	return 0;
}

sub _findFiles($$$);
sub _findFiles($$$) {
	my $dir = shift;                   # directory as given by user
	my $triggers = shift;
	my $fullCanonizedBaseDir = shift;  # same directory but absolute and normalized

	my $ret = opendir DIR, $dir;

	if (! $ret) {
		print "unable to read directory : $dir\n";
		return
	}
	my @entries = readdir DIR;
	closedir DIR;

	for my $file (@entries) {

		next if (isIgnored($file));

		my $fullfilename = $dir . "/" . $file;
		my $fullCanonizedBasePath = $fullCanonizedBaseDir."/".$file;
#print STDERR "FULL CANONIZED BASE PATH = $fullCanonizedBasePath\n";
		if (isIgnoredPath($fullCanonizedBasePath)) {
#print STDERR " ---> IGNORE $fullCanonizedBasePath\n";
			next;
		}

		if (Lib::LongPathUtils::IsFile($fullfilename)) {
			#if ($file eq $findFiles_filePattern) {
			if (findFiles_fileMatch($file)) {
				#push @$findFiles_fileList, $fullfilename;
				for my $trigger (@$triggers) {
					$trigger->($fullfilename);
				}
			}
		}
		elsif (Lib::LongPathUtils::IsFolder($fullfilename)) {
			if (($file eq ".") || ($file eq "..")) {
				next;
			}
			_findFiles($fullfilename, $triggers, $fullCanonizedBasePath);
		}
	}
}

sub findFiles($$;$) {
	my $dir = shift;
	$findFiles_filePattern = shift;
	my $callbacks = shift;

	my $fullCanonizedBaseDir = "";
	if (! isAbsolutePath($dir)) {
		if ( $ROOT_DIR ne getcwd()) {
			print STDERR "[Lib::Sources::findFiles] ERROR : working directory ".getcwd()." has changed (was $ROOT_DIR at beginning of the analysis). Relative path $dir may be erroneous !\n"; 
		}
	}

	$fullCanonizedBaseDir = buildAbsoluteCanonizedPath($dir);

    if (! -d $fullCanonizedBaseDir) {
		print STDERR "[Lib::Sources::findFiles] ERROR : directory $dir is canonized to $fullCanonizedBaseDir, but this doesn't exist!!\n"; 
	}

	if (! defined $callbacks) {
		$callbacks = [\&findFiles_triggerAdd];

		# Init the output list.
		$findFiles_fileList = [];

		# search ...
		_findFiles($dir, $callbacks, $fullCanonizedBaseDir);

		return $findFiles_fileList;
	}

	_findFiles($dir, $callbacks, $fullCanonizedBaseDir);
	return 0;
}

########################################################################

my @T_Triggers = ();

sub registerPostAnalysisProjectFilesTrigger($) {
	my $callback = shift;
	push @T_Triggers, $callback;
}

sub postAnalysisProjectFilesScanning($) {
	my $tabSourcesDir = shift;
	if (scalar @T_Triggers) {
		findFiles($tabSourcesDir->[0], undef, \@T_Triggers);
	}
}

########################################################################
#                  Check encoding
########################################################################

my @encoding_list = ('iso_8859-1', 'windows-1252');

sub needEncodingForExisting($$) {
	my $path = shift;
	my $filename = shift;
	if (!Lib::LongPathUtils::PathExists($path . $filename)) {
		# the file cannot be found  ==> try different enconding
		for my $encoding (@encoding_list) {
			my $encoded = Encode::encode($encoding, $filename);
			if (Lib::LongPathUtils::PathExists($path . $encoded)) {
				return $encoding;
			}
		}
	}
	else {
		# the file can be found ==> needed encoding is none
		return undef;
	}

	return "unknow";
}

1;
